package strickling.utils;

/* Simple class for displaying a camera preview or blinking torch on SGS2
 * 
 */

/*****************************************************************************\
 *
 *  Copyright 2019 Dr. Wolfgang Strickling
 *  published under  GNU GPL 3
 *
 *   This file is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 \*****************************************************************************/
 



import java.io.IOException;
import android.annotation.SuppressLint;
import android.annotation.TargetApi;
import android.content.Context;
import android.hardware.Camera;
import android.os.Build;
import android.util.Log;
import android.view.SurfaceHolder;
import android.view.SurfaceView;

@TargetApi(Build.VERSION_CODES.FROYO)
public class ARView extends SurfaceView implements SurfaceHolder.Callback {

	public static String TAG = "ARView";

	public final static int CAMERA_ERROR_NO_INIT = -1;
	public final static int CAMERA_ERROR_NO_CAM  = 1;
	public final static int CAMERA_ERROR_IO_ERR  = 2;
	public final static int CAMERA_PREVIEW_OK  	 = 21;

	public static int    	cameraError = CAMERA_ERROR_NO_INIT;
	public static int	 	blinkFlashMillis	 = 0;


	public SurfaceHolder mHolder;  // <2>
	public Camera camera; // <3>
	public static double  horizontalViewAngle = 52; // may be overwritten
	public static double  verticalViewAngle = horizontalViewAngle *480.0/800.0; // dummy, will be recalculated
    public static boolean autoViewAngle;		// if false, get manual values from input / Settings in ARActivity Class 


	public ARView(Context context) {
		super(context);
		// Install a SurfaceHolder.Callback so we get notified when the
		// underlying surface is created and destroyed.
		mHolder = getHolder();  // <4>
		mHolder.addCallback(this);  // <5>
		mHolder.setType(SurfaceHolder.SURFACE_TYPE_PUSH_BUFFERS); // <6>
	}
	
	private Camera.Size getBestPreviewSize(int width, int height)
	{ // from http://stackoverflow.com/questions/5540981/picture-distorted-with-camera-and-getoptimalpreviewsize
	        Camera.Size result=null;    
	        Camera.Parameters p = camera.getParameters();
	        for (Camera.Size size : p.getSupportedPreviewSizes()) {
	            if (size.width<=width && size.height<=height) {
	                if (result==null) {
	                    result=size;
	                } else {
	                    int resultArea=result.width*result.height;
	                    int newArea=size.width*size.height;

	                    if (newArea>resultArea) {
	                        result=size;
	                    }
	                }
	            }
	        }
	    return result;

	}


	@SuppressLint("NewApi")
	public void surfaceChanged(SurfaceHolder holder, int format, int width, int height) {
		// The Surface has been created, acquire the camera and tell it where
		// to draw.
		Log.d (TAG, "CameraPreview: Try Camera.open()");
		try {
			cameraError = CAMERA_ERROR_NO_CAM;
			camera = Camera.open(); // <8>
			Log.d (TAG, "CameraPreview: Camera.open() oK");
			try {
			    Camera.Parameters parameters = camera.getParameters();
			    Camera.Size size = getBestPreviewSize(width, height);
			    if (size != null) {
			    	parameters.setPreviewSize(size.width, size.height);
			    	camera.setParameters(parameters);}			  

				camera.setPreviewDisplay(holder);  // <9>
				camera.startPreview();
				cameraError = CAMERA_PREVIEW_OK;
				
				verticalViewAngle = horizontalViewAngle
				  *camera.getParameters().getPreviewSize().width
				  /camera.getParameters().getPreviewSize().height;

				//Camera.Parameters parameters = camera.getParameters();
				if (blinkFlashMillis > 0){ // switch off flash
					Log.d(TAG, "CameraActivity Blink Torch: StartFlash " +blinkFlashMillis +" ms");
					try {
						parameters.setAutoExposureLock(true); // here App crash
						parameters.setFocusMode(Camera.Parameters.FOCUS_MODE_INFINITY);
					} catch (Exception e) {}
					parameters.setFlashMode(Camera.Parameters.FLASH_MODE_TORCH); //TODO
					camera.setParameters(parameters);
					try {
						Thread.sleep (blinkFlashMillis); //TODO may produce ANR here, put to separate thread
						Log.d(TAG, "CameraActivity Blink Torch: StopFlash");
					} catch (InterruptedException e) {			
						e.printStackTrace();
						Log.d(TAG, "CameraActivity Blink Torch: SleepError");
						//CameraPreview.errorLog += "\n" +"CameraActivity Blink Torch: SleepError:" +"\n" +Log.getStackTraceString(e);
					}
					destroy();
					blinkFlashMillis = 0;
				}
				if (autoViewAngle && (Build.VERSION.SDK_INT >= 8)){ // only Ver 2.2 or higher
					Log.d(TAG, "ARView.srfChanged: read Auto FoV");
					horizontalViewAngle = parameters.getHorizontalViewAngle(); 
					verticalViewAngle   = parameters.getVerticalViewAngle();
				}
				Log.d(TAG, String.format("ARView.srfChanged: %.0f deg x %.0f deg Auto: "  +autoViewAngle, horizontalViewAngle, verticalViewAngle));
			} catch (IOException e) { // <13>
				e.printStackTrace();
				Log.d (TAG, "CameraPreview: IOException");
			}
		} //finally {} //Log.d(TAG, "onPreviewFrame Camera open error");}
		catch (RuntimeException e) { // <13>
			e.printStackTrace();
			camera = null;
			Log.d (TAG, "CameraPreview: RuntimeException " +e.toString());
		}
	}

	public void surfaceCreated(SurfaceHolder holder) {
	}

	public void surfaceDestroyed(SurfaceHolder holder) {
		destroy();
	}
	
	public void destroy (){
		Log.d(TAG, "ARView.surfaceDestroyed");
		if (camera != null) try {			
			if (blinkFlashMillis > 0){
				Camera.Parameters parameters = camera.getParameters();  //TODO
				parameters.setFlashMode(Camera.Parameters.FLASH_MODE_OFF); 
				camera.setParameters(parameters);
			}
			
			camera.setPreviewCallback(null); // inserted after Heinz Crash report, see
			//   http://stackoverflow.com/questions/3592283/camera-force-close-when-returning-from-sleep-android
			camera.stopPreview();	  		
			//mHolder.removeCallback(this);
			camera.release(); // releases the camera object. Does NOT take picture! Otherwise next Camera.open will crash
			camera=null;     // der muss hier sein
		} catch (NullPointerException e) {
			e.printStackTrace();
		} catch (RuntimeException e) {
			e.printStackTrace();
		}
		
	}
}
