package strickling.utils;

/*****************************************************************************\
 *
 *  Copyright 2019 Dr. Wolfgang Strickling
 *  published under  GNU GPL 3
 *
 *   This file is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 \*****************************************************************************/
 
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Locale;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import android.util.Log;

/* TimeZone from Coordinates: 
 * https://developers.google.com/maps/documentation/timezone/
 * http://developer.yahoo.com/geo/placefinder/guide/requests.html
 * (from http://stackoverflow.com/questions/41504/timezone-lookup-from-latitude-longitude/41533#41533)
 * 
 */

public class DateTimeUtils {
	
	public static final double JD1950       = 2433282.423;  //JulDat von B1950.0
	public static final double JD2000       = 2451545.0  ;  //JulDat von J2000.0
	public static final double JD_ROUND_PREVENTER  = 0.00000001;
	public static final String TAG 			= "DateTimeUtils";
	public static final double JD_MJD0	    = 2400000.5;
	public static final String DEFAULT_DATE_FORM = "yyyy - MM - dd";
	public static final String DEFAULT_DATETIME_SEPARATOR = "  ";
	public static final String DEFAULT_TIME_FORM = "HH:mm:ss";
	public static final String DEFAULT_DATETIME_FORM = DEFAULT_DATE_FORM +DEFAULT_DATETIME_SEPARATOR +DEFAULT_TIME_FORM;
	
	/**
	 * returns String "yyyy - MM - dd" from julDat
	 * @param julDat  
	 * @param timeZoneHrs in hours
	 */
	public static String YMDStrg (double julDat, double timeZoneHrs) {
// 		AstroDate ad = new AstroDate (julDat + timeZone/24. +0.5 +JD_ROUND_PREVENTER);
// 		return String.format("%02d - %02d - %02d",   ad.year(), ad.month(), ad.day());
		double [] ymd = new double [3];
		gregDatCalc (julDat, timeZoneHrs, ymd);
 		return String.format("%02.0f - %02.0f - %02d", ymd [0], ymd [1], (int) ymd [2]);
		}
	
	/**returns String "h:mm:ss" from duration in days,  
	 *  
	 * @param days  
	 * @param plusSign  if true give leading "+"
	 * @return String "h:mm:ss" from days
	 */
	public static String HMSDurStrg (long millis, boolean plusSign) {
		return HMSDurStrg (millis/(86400000.0), plusSign);
	}
	public static String HMSDurStrg (Locale locale, long millis, boolean plusSign) {
		return HMSDurStrg (locale, millis/(86400000.0), plusSign);
	}
		
	/**returns String "h:mm:ss" from duration in days,  
	 *  
	 * @param days  
	 * @param plusSign  if true give leading "+"
	 * @return String "h:mm:ss" from days
	 */
	public static String HMSDurStrg (double days, boolean plusSign) {
		return HMSDurStrg (Locale.getDefault(), days, plusSign);
	}
	
	public static String HMSDurStrg (Locale locale, double days, boolean plusSign) {
		String s = "";
		if (days <0) {s = "-";}
		else if (plusSign) { s = "+"; }
		
		days = (Math.abs(days) + 0.5/86400.0) *24.0;
		int hour = (int) days;
		days = (days -hour) *60.;
		int minute = (int) days;
		days = (days -minute) *60.;
		int second = (int) Math.round(days-0.5);
		return String.format(locale, s+"%02d:%02d:%02d",   hour, minute, second);
	}
	
	/**  returns String linke "h:mm:ss" from days if days > 0, else "--:--:--"
	 * 
	 * @param days
	 * @return
	 */
	public static String formatDays (double days) {
		if (days >= 0) { return HMSDurStrg (days, false);}
		else return "--:--:--";
	}

	/**returns String "mm:ss" from duration in minutes,  
	 *  
	 * @param durMinutes 
	 * @param plusSign
	 * @param secDecimals  
	 * @return String "h:mm:ss" from julDat
	 */
	public static String MSDurStrg (double durMinutes, boolean plusSign, int secDecimals) {
		return MSDurStrg (Locale.getDefault(), durMinutes, plusSign, true, secDecimals);
	}
		
	public static String MSDurStrg (Locale locale, double durMinutes, boolean plusSign, int secDecimals) {
		return MSDurStrg (locale, durMinutes, plusSign, true, secDecimals);
	}
		
		/**returns String "mm:ss" from duration in minutes,  
		 *  
		 * @param durMinutes 
		 * @param plusSign
		 * @param leadingZero
		 * @param secDecimals  
		 * @return String "h:mm:ss" from julDat
		 */
	public static String MSDurStrg (Locale locale, double durMinutes, boolean plusSign, boolean leadingZero, int secDecimals) {
		
		String s = "";
		if (durMinutes <0) {s = "-";}
		else if (plusSign) { s = "+"; }
		
		durMinutes = (Math.abs(durMinutes));
		int minute = (int) durMinutes;
		double second = (durMinutes -minute) *60.;
		
		//avoid round errors like "3:60.0"
		if (String.format ("%."+secDecimals+"f", second).startsWith("60", 0)){
			second = 0;
			minute ++;
		}
		return String.format(locale, s+"%" +(leadingZero ? "0" : "") +"2d:%0" +(secDecimals == 0 ? 2 : 3+secDecimals) +"."+secDecimals+"f",   minute, second);
	}

	/**returns String "h:mm:ss" from julDat
	 *  
	 * @param julDat  
	 * @param timeZoneHrs in hours
	 * @return String "h:mm:ss" from julDat
	 */
	public static String HMSZeitStrg (double julDat, double timeZoneHrs) {
		
//		AstroDate ad = new AstroDate (julDat + timeZone/24.0 +JD_ROUND_PREVENTER);
//		return String.format("%02d:%02d:%02d",   ad.hour(), ad.minute(), ad.second());
		return HMSZeitStrg (julDat, timeZoneHrs, 0);
	}

	
	/**returns String "hh:mm:ss.s" from julDat
	 *  
	 * @param julDat  
	 * @param timeZoneHrs in hours
	 * @return String "hh:mm:ss" from julDat
	 */
	public static String HMSZeitStrg (double julDat, double timeZoneHrs, int secDecimals) {
		return HMSZeitStrg (Locale.getDefault(), julDat, timeZoneHrs, secDecimals);
	}
	public static String HMSZeitStrg (Locale locale, double julDat, double timeZoneHrs, int secDecimals) {

		double [] hms = new double [3];
		days2HMS (AstroUtils.frac(julDat +0.5 +JD_ROUND_PREVENTER +timeZoneHrs/24.0), secDecimals, hms);
		
		String secForm = "%0"+(secDecimals+3)+"."+secDecimals+"f";
		if (secDecimals == 0) secForm = "%02.0f"; 
		return String.format(locale, "%02d:%02d:" +secForm, (int) hms[0], (int) hms[1], hms[2]);
//		AstroDate ad = new AstroDate (julDat + timeZone/24 +JD_ROUND_PREVENTER);
//		String s = "%02d:%02d:%0"+(secDecimals+3)+"."+secDecimals+"f";
//		return String.format("%02d:%02d:%0"+(secDecimals+3)+"."+secDecimals+"f",   ad.hour(), ad.minute(), (double) ad.second());
	}

	
	/**returns String "hh:mm" from julDat
	 *  
	 * @param julDat  
	 * @param timeZoneHrs in hours
	 * @return String "hh:mm" from julDat
	 */
	public static String HMZeitStrg (double julDat, double timeZoneHrs) {
		double [] hms = new double [3];		
		// round to Minutes
		double days = AstroUtils.roundFloat(
				AstroUtils.frac(julDat +0.5 +JD_ROUND_PREVENTER +timeZoneHrs/24.0),
				1.0/(24.0*60.0));
		days2HMS (days, 0, hms);
		return String.format("%02.0f:%02.0f", hms[0], hms[1]+ hms[2]/60.0);

	}

	/**returns String "yyyy - mm - dd  h:mm:ss" from now with current timeZone
	 *  
	 * @param julDat  
	 * @param timeZone in hours
	 * @return String "yyyy - mm - dd  h:mm:ss" from julDat
	 */
	public static String zeitStrg () {
		return zeitStrg (nowJulDat (), GetCurrentTimeZone());
	}
		/**returns String "yyyy - mm - dd  h:mm:ss" from julDat
		 *  
		 * @param julDat  
		 * @param timeZoneHrs in hours
		 * @return String "yyyy - mm - dd  h:mm:ss" from julDat
		 */
	public static String zeitStrg (double julDat, double timeZoneHrs) {
//		AstroDate ad = new AstroDate (julDat + timeZone/24 +JD_ROUND_PREVENTER);
//		return String.format("%02d - %02d - %02d  %d:%02d:%02d",   
//				 ad.year(), ad.month(), ad.day(), ad.hour(), ad.minute(), ad.second());
		double [] t = new double [6];
		jd2YMDHMS (julDat, timeZoneHrs, 0, t);
		return String.format("%02.0f - %02.0f - %02.0f  %02.0f:%02.0f:%02.0f",   
				t [0], t [1], t [2], t [3], t [4], t [5]);
	}
	
  	/**
	 * returns String like "UTC +03.5 h"
	 */
	public static String UTCZeitZoneStrg (double timeZoneHrs) {
		return String.format("UTC %+04.1f h", timeZoneHrs);
	}
	
	/** returns DST corrected offset of TimeZone for julDat
	 * 
	 * @param julDat
	 * @param rawTimeZoneHrs
	 * @return
	 */
	public static double getTimeZoneHrs (double julDat, double rawTimeZoneHrs) {
		TimeZone tz = TimeZone.getTimeZone("UTC");
		tz.setRawOffset((int) (rawTimeZoneHrs*3600000)); //Log.d(TAG, "    getTimeZone: " +tz.getRawOffset());
    	return tz.getOffset(DateTimeUtils.julDat2Millis(julDat))/ 3600000.;
	}

	/** returns DST corrected offset of TimeZone for julDat
	 * 
	 * @param julDat
	 * @param rawTimeZoneHrs
	 * @return
	 */
	public static double getTimeZoneHrs (double julDat, TimeZone tz) {
    	return tz.getOffset(DateTimeUtils.julDat2Millis(julDat))/3600000.;
	}

	
	
	
	/** system's time zone offset (hours)
	 * 	
	 * @return Time Zone (TZ + DST) offset in Hours
	 */
	public static double GetCurrentTimeZone () {
      	TimeZone tz = TimeZone.getDefault();
		return (tz.getRawOffset()/3600000.0) +GetCurrentDSTOffset ();		
		}
	

	/**
	 * 	
	 * @return DST offset in Hours
	 */
	public static double GetCurrentDSTOffset () {
      	TimeZone tz = TimeZone.getDefault();
     	Date dt = new Date ();
     	if (tz.inDaylightTime (dt)) {return tz.getDSTSavings()/3600000.0;}
     	else return 0;
		}

	/** gets TimeZone object WITHOUT DST, eg. name for CEST is GMT+02:00 
	 * 
	 * @param timeZoneHrs
	 * @return
	 */
	public static TimeZone getTimeZone (double timeZoneHrs) {
		TimeZone tz = TimeZone.getTimeZone("GMT" +DateTimeUtils.MSDurStrg(Locale.getDefault(), timeZoneHrs, true, true, 0));
//		tz.setRawOffset((int) (timeZoneHrs*3600000)); //Log.d(TAG, "    getTimeZone: " +tz.getRawOffset());
		return tz;
	}
	
	
	public static double nowJulDat () {
		GregorianCalendar cal = new GregorianCalendar();
		return Cal2JulDat (cal);
	}
	
	/** gets julian date from year
	 * 
	 * @param epochYear, eg. 2000.0
	 * @return Julian Date
	 */
	public static double epoch2JulDat (double epochYear) {
		return (epochYear -2000.0)*365.25 +JD2000;		
	}
	
	/** gets julian date from Java long Time format in Milliseconds
	 * 
	 * @param millis
	 * @return Julian Date
	 */
	public static double millis2JulDat (long millis) {
		return 2440587.5 +(millis)/(86400.0*1000.0);		
	}
	
	/** gets julian date from Java long Time format in Milliseconds
	 * 
	 * @param millis
	 * @return Julian Date
	 */
	public static long julDat2Millis (double jd) {
		return (long) ((jd-2440587.5) *(86400.0*1000.0));		
	}
	
	public static Calendar julDat2Cal (Locale locale, double julDat) {
    	Calendar cal = Calendar.getInstance(TimeZone.getDefault(), locale);
    	cal.setTimeInMillis(DateTimeUtils.julDat2Millis(julDat));		
		return cal;		
	}

	
	public static double Cal2JulDat (GregorianCalendar cal) {
		return jdCalc (cal.get(Calendar.YEAR),
				cal.get(Calendar.MONTH)+1,
				cal.get(Calendar.DATE)
				+(cal.get(Calendar.HOUR_OF_DAY))/24.0
				+cal.get(Calendar.MINUTE)/(24.0*60.0)
				+cal.get(Calendar.SECOND)/86400.0
				+cal.get(Calendar.MILLISECOND)/86400000.0,
				GetCurrentTimeZone());
	}

	public static double jdCalc (int y, int m, int d, int h, int min, double sec, double timeZone) {
		return jdCalc (y, m, d +h/24.0 +min/(24.0*60.0) +sec/86400.0, timeZone);
	}

	/** Converts Gregorian/Julian date to astronomic JulianDate
	 * 
	 * @param y
	 * @param m
	 * @param d
	 * @param timeZone East in hours
	 * @return
	 */
	public static double jdCalc (int y, int m, double d, double timeZoneHrs) {
		// Juldat nach Montenbruck S. 50 , UT
		double Hilfsgr;
		int B;

		if (m<=2){
			y = y-1;
			m = m+12;
		}
		if ((y +m/100 +d/1000) >= 1582.1015)  //after der Greg. Kalenderreform
		{B = y/400 - y/100;}
		else B = -2;

		if (y >0)
		{Hilfsgr = Math.floor (365.25*y);}
		else Hilfsgr = Math.floor (365.25*y -0.75);

		return Hilfsgr  +Math.floor(30.6001 *(m+1))  +B  +d  +1720996.5 -timeZoneHrs/24.0;
	} // of JDcalc 
	
	/** Converts Gregorian/Julian date to astronomic JulianDate,
	 * 
	 * @param String s format "yyyy - MM - dd  hh:mm:ss"
	 * @param timeZoneHrs East in hours
	 * @return Julian Date
	 */
	   public static double jdCalc (String s, double timeZoneHrs) {
	    	int y = -9999;
	    	int mon = 0;
	    	int d = 0;
	    	int h = 0;
	    	int min = 0;
	    	int sec = 0;
	    	Pattern p = Pattern.compile("[0-9]+");
	    	Matcher m = p.matcher(s);
	    	if (m.find()) {y = Integer.parseInt(m.group());}
	    	if (m.find()) {mon = Integer.parseInt(m.group());}
	    	if (m.find()) {d = Integer.parseInt(m.group());}
	    	if (mon == 0 || d == 0) return DateTimeUtils.jdCalc (-9999, 9, 9.9999, timeZoneHrs);
	    	if (m.find()) {h = Integer.parseInt(m.group());}
	    	if (m.find()) {min = Integer.parseInt(m.group());}
	    	if (h < 0 || h >= 24 || min < 0 || min >= 60) return DateTimeUtils.jdCalc (y, mon, d, timeZoneHrs);
	    	if (m.find()) {sec = Integer.parseInt(m.group());}
	    	if (sec < 0 || sec >= 60) return DateTimeUtils.jdCalc (y, mon, d, h, min, 0, timeZoneHrs);
	//    	Log.d(TAG, String.format("found %d - %d - %d  %d:%d:%d", y, mon, d, h, min, sec));
	    	return DateTimeUtils.jdCalc (y, mon, d, h, min, sec, timeZoneHrs);
	    }

	/** Converts astronomic JulianDate to Gregorian/Julian date
	 * 
	 * @param julDat
	 * @param timeZoneHrs East in hours
	 * @param ymd 0: Year, 1: Month, 2: Day
	 */
	public static void gregDatCalc (double julDat, double timeZoneHrs, double [] ymd){
		double a, b, c, d, e, f;
		julDat += timeZoneHrs/24.0;
		a = Math.floor(julDat +0.5 +JD_ROUND_PREVENTER);
		if (a < 2299161.0)
		{ c = a+1524.0;}
		else {
			b = Math.floor((a -1867216.25)/36524.25);
			c = a +b -Math.floor(b/4) +1525;
		}
		d = Math.floor((c -122.1)/365.25);
		e = Math.floor(365.25 *d);
		f = Math.floor((c-e)/30.6001);

		ymd [2] = Math.round (c -e -Math.floor(30.6001*f)) +AstroUtils.frac(julDat +0.500001);  // day
		ymd [1] = Math.round (f -1.0 -12.0*Math.floor(f/14.0));                     // month
		ymd [0] = Math.round (d -4715.0 -Math.floor((7.0+ymd [1]) /10.0) );         // year
	} // of gregDatCalc

	public static int gregDatYear (double julDat, double timeZoneHrs) {
		double [] ymd = new double [3]; // set year to Eclipse in EclipseCalculator
		DateTimeUtils.gregDatCalc(julDat, 0, ymd);
		return (int) ymd [0];
	}

	/**
	 * 
	 * @param days
	 * @param secondDecimals for rounding purpose
	 * @param hms 0: hours, 1: minutes, 2: seconds
	 */
	public static void days2HMS (double days, int secondDecimals, double[] hms){
		double rounder = 1.0/Math.pow(10.0, secondDecimals);
		
		double a = (Math.abs(days) + 0.5*rounder/86400.0) *24.0;
		int hour = (int) a;
		a = (a -hour) *60.;
		int minute = (int) a;
		a = (a -minute) *60.;
		hms[2] = Math.round((a-0.5*rounder)/rounder)*rounder; // s
		hms[1] = minute;
		hms[0] = hour * Math.signum (days);
	}
	
	/** Converts astronomic JulianDate to Gregorian/Julian date
	 * 
	 * @param days
	 * @param secondDecimals
	 * @param ymdhms 0: Year, 1: Month, 2: Day, 3: hours, 4: minutes, 5: seconds
	 */
	public static void jd2YMDHMS (double julDat, double timeZoneHrs, int secondDecimals, double[] ymdhms){
		gregDatCalc (julDat, timeZoneHrs, ymdhms);
		
		double [] hms = new double[3];
		days2HMS (AstroUtils.frac(ymdhms[2]), secondDecimals, hms);
		ymdhms[2] = Math.floor(ymdhms[2]);
		System.arraycopy(hms, 0, ymdhms, 3, 3);
	}
	
	/** conversion Julian Date into Mod. julian Date (for Montenbruck-Pfleger routines(
	 * 
	 * @param julDat
	 * @returns MJD = JD-2400000.5
	 */
	public static double jd2MJD (double julDat) {
		return julDat -JD_MJD0;		
	}
	
	public static double mjd2JD (double mjd) {
		return mjd +JD_MJD0;		
	}
	
	/** Converts long date line 20110123 into julian day UTC
	 * 
	 * @param yyyymmdd
	 * @return
	 */
	public static double ymd2JD (long yyyymmdd) {
		 long lyear  = yyyymmdd / 10000;
		 long lmonth = (yyyymmdd % 10000) / 100;
		 long lday   = yyyymmdd % 100;
		 int  year   = (int) lyear;
		 int  month  = (int) lmonth;
		 double day  = (int) lday;

		 return jdCalc(year, month, day, 0.0);
	}
	
	/** Number of day of Week [Mon = 0... Sun = 6]
	 * 
	 * @param julDat
	 * @param timeZoneHrs in hours
	 * @return day of Week number [Mon = 0... Sun = 6]
	 * usage example:
	 
	 Resources res = getResources();
     CharSequence[] daysArray = res.getTextArray(R.array.ShortDays);
     
     // Get the string:
     (String) daysArray[DateTimeUtils.dayOfWeek(julDat, timeZone)]
	 
	 */
	public static int dayOfWeek (double julDat, double timeZoneHrs) {
		double jd_modif = Math.round (julDat +timeZoneHrs/24.0 +JD_ROUND_PREVENTER);
		return (int) Math.round ((jd_modif/ 7 -Math.floor (jd_modif/ 7)) *7);		
	}

//	public static String dayOfWeek (double julDat, double timeZone, int RESOURCE_ID) {
//		 Resources res = getResources();
//	     CharSequence[] daysArray = res.getTextArray(RESOURCE_ID);
//		return (String) daysArray[DateTimeUtils.dayOfWeek(julDat, timeZone)];
//	}

	/** Calculates Delta UTC | DeltaUTC=TDT-UTC |  TDT = UTC + DeltaUTC | UTC = TDT -DeltaUTC
	 * 
	 * @param jd Julian Date
	 * @return Delta UTC in days, valid form -500 to 2050
	 */
	public static double CalcDeltaUTC    (double jd) {
		if (jd >= 2441317.5) { // 2970-01-01 was introduction of leap seconds
			double deltaUTC = 42.184;
			if (jd >= 2441499.5) {deltaUTC ++;} // 1.7.1972
			if (jd >= 2441683.5) {deltaUTC ++;} // 1.1.1973
			if (jd >= 2442048.5) {deltaUTC ++;} // 1.1.1974
			if (jd >= 2442413.5) {deltaUTC ++;} // 1.1.1975
			if (jd >= 2442778.5) {deltaUTC ++;} // 1.1.1976
			if (jd >= 2443144.5) {deltaUTC ++;} // 1.1.1977
			if (jd >= 2443509.5) {deltaUTC ++;} // 1.1.1978
			if (jd >= 2443874.5) {deltaUTC ++;} // 1.1.1979
			if (jd >= 2444239.5) {deltaUTC ++;} // 1.1.1980
			if (jd >= 2444786.5) {deltaUTC ++;} // 1.7.1981
			if (jd >= 2445151.5) {deltaUTC ++;} // 1.7.1982
			if (jd >= 2445516.5) {deltaUTC ++;} // 1.7.1983
			if (jd >= 2446247.5) {deltaUTC ++;} // 1.7.1985
			if (jd >= 2447161.5) {deltaUTC ++;} // 1.1.1988
			if (jd >= 2447892.5) {deltaUTC ++;} // 1.1.1990
			if (jd >= 2448257.5) {deltaUTC ++;} // 1.1.1991
			if (jd >= 2448804.5) {deltaUTC ++;} // 1.7.1992
			if (jd >= 2449169.5) {deltaUTC ++;} // 1.7.1993
			if (jd >= 2449534.5) {deltaUTC ++;} // 1.7.1994
			if (jd >= 2450083.5) {deltaUTC ++;} // 1.1.1996
			if (jd >= 2450630.5) {deltaUTC ++;} // 1.7.1997
			if (jd >= 2451179.5) {deltaUTC ++;} // 1.1.1999
			if (jd >= 2453736.5) {deltaUTC ++;} // 1.1.2006
			if (jd >= 2454466.5) {deltaUTC ++;} // 1.1.2008
			if (jd >= 2455927.5) {// 1.1.1012
				return estimateDeltaUTC (CalcDeltaT(jd));}
			else return deltaUTC/86400.0;
			}
		else return CalcDeltaT(jd);
	}

	
	/** estimates deltaUTC in days
	 * @param deltaT in days
	 */
	public static double estimateDeltaUTC (double deltaT){
		return (Math.floor (deltaT*86400.0 +0.4 -0.184)+0.184)/86400.0;
	}
	
	
	/** Calculates Delta T  in DAYS | DeltaT=TDT-UT |  TDT = UT + DeltaT | UT = TDT -DeltaT
	 * 
	 * http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html 
	 * @param jd Julian Date
	 * @return Delta T in days, valid form -500 to 2050
	 */
	public static double CalcDeltaT    (double jd) {
		/*   neuer Algorithmus nach http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html
		 *   DeltaT=ET-UT1, UT1 = ET -DeltaT, TDT = UTC + DeltaT
		 *   users may compute Delta T using the relation:
     			Delta T = 32.184 s + (TAI - UTC) - (UT1 - UTC)
				where
     			(TAI - UTC) is the cumulative number of leap seconds (=34.000 s since 2009-01-01), changes only at leap seconds
				and
     			(UT1 - UTC) is a standard EOP product available below. (if > 0.9 s => leap second).
     			   changes continuously, if leap second added instant, http://de.wikipedia.org/wiki/Schaltsekunde
		 *   
		 *   
		 */
		double 
		u, t,        // auxiliary params,
		y,			 // Year
		DeltaT;      // in s


		y = jd/365.2425 -4712.0991;
		DeltaT = 0;

		if (y <-500) {
			u      = (y-1820)/100;
			DeltaT = -20 + 32 * u*u;
		}
        if ((y >=-500) && (y<500)) {
			u      = y/100;
			DeltaT = 10583.6 - 1014.41 *u + 33.78311 * u*u - 5.952053 *u*u*u
			- 0.1798452 *u*u*u*u + 0.022174192 * u*u*u*u*u + 0.0090316521 *u*u*u*u*u*u;
		}
        if ((y >=500) && (y<1600)) {
			u      = (y-1000)/100;
			DeltaT = 1574.2 - 556.01 * u + 71.23472 * u*u + 0.319781 * u*u*u
			- 0.8503463 *u*u*u*u - 0.005050998 *u*u*u*u*u + 0.0083572073 * u*u*u*u*u*u;
		}
        if ((y >=1600) && (y<1700)) {
			t      = y - 1600;
			DeltaT = 120 - 0.9808 * t - 0.01532 * t*t + t*t*t / 7129;
		}
        if ((y >=1700) && (y<1800)) {
			t      = y - 1700;
			DeltaT = 8.83 + 0.1603 * t - 0.0059285 * t*t + 0.00013336 * t*t*t - t*t*t*t / 1174000;
		}
        if ((y >=1800) && (y<1860)) {
			t      = y - 1800;
			DeltaT = 13.72 - 0.332447 * t + 0.0068612 * t*t + 0.0041116 * t*t*t - 0.00037436 * t*t*t*t
			+ 0.0000121272 * t*t*t*t*t - 0.0000001699 * t*t*t*t*t*t + 0.000000000875 * t*t*t*t*t*t*t;
		}
        if ((y >=1860) && (y<1900)) {
			t      = y - 1860;
			DeltaT = 7.62 + 0.5737 * t - 0.251754 * t*t + 0.01680668 * t*t*t
			-0.0004473624 * t*t*t*t + t*t*t*t*t / 233174;
		}
        if ((y >=1900) && (y<1920)) {
			t      = y - 1900;
			DeltaT = -2.79 + 1.494119 * t - 0.0598939 * t*t + 0.0061966 * t*t*t - 0.000197 * t*t*t*t;
		}
        if ((y >=1920) && (y<1941)) {
			t      = y - 1920;
			DeltaT = 21.20 + 0.84493*t - 0.076100 * t*t + 0.0020936 * t*t*t;
		}
        if ((y >=1941) && (y<1960)) {
			t      = y - 1950;
			DeltaT = 29.07 + 0.407*t - t*t/233 + t*t*t / 2547;
		}
        if ((y >=1961) && (y<1986)) {
			t      = y - 1975;
			DeltaT = 45.45 + 1.067*t - t*t/260 - t*t*t / 718;
		}
        if ((y >=1986) && (y<2005)) {
			t      = y - 2000;
			DeltaT = 63.86 + 0.3345 * t - 0.060374 * t*t + 0.0017275 * t*t*t + 0.000651814 * t*t*t*t
			+ 0.00002373599 * t*t*t*t*t;
		}
        if ((y >=2005) && (y<2050)) {
			t      = y - 2000;
			DeltaT = 62.92 + 0.32217 * t + 0.005589 * t*t;
		}
        if ((y >=2050) && (y<2150)) {
			DeltaT = -20 + 32 * Math.sqrt((y-1820)/100) - 0.5628 * (2150 - y);
		}
        if (y >=2150) {
			u      = (y-1820)/100;
			DeltaT = -20 + 32 * u*u;
		}

        return DeltaT / 86400.0;
	} //CalcDeltaT

}
