package strickling.utils;

/*****************************************************************************\
 *
 *  Copyright 2019 Dr. Wolfgang Strickling
 *  published under  GNU GPL 3
 *
 *   This file is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 \*****************************************************************************/
 

import com.mhuss.AstroLib.AstroOps;
import com.mhuss.AstroLib.LunarCalc;
import com.mhuss.AstroLib.MathOps;
import com.mhuss.AstroLib.NoInitException;
import com.mhuss.AstroLib.Nutation;
import com.mhuss.AstroLib.ObsInfo;
import com.mhuss.AstroLib.PlanetData;
import com.mhuss.AstroLib.Planets;
import android.util.Log;

public class EphemUtils {
	
	public static final String TAG = "EphemUtils";
	
	public static final double TWO_PI  = Math.PI *2.0;
	public static final double HALF_PI = Math.PI /2.0;
	public static final double RADFAKT = Math.PI /180.0;
	public static final double DEGFAKT = 180.0/Math.PI;


	/******************************* konvertierte Pascal-Sourcen **************/
	/*---------------------------------------------------------------------------*/
	/*  STUMPFF: Berechnung der Stumpff-Funktionen  C1 = Math.sin(E)/E ,         */
	/*           C2 = (1-Math.cos(E))/(E**2) und C3 = (E-Math.sin(E))/(E**3)     */
	/*           zum Argument E2=E**2                                            */
	/*           (E: exzentrische Anomalie im Bogenmass)                         */
	/*---------------------------------------------------------------------------*/
	public static void stumpff (double E2, double [] c) {
		final double EPS=1E-12;
		double N,ADD;

		c[0]=0.0; c[1] =0.0; c[2]=0.0; 
		ADD=1.0; N=1.0;
		do {
			c[0]=c[0]+ADD; ADD=ADD/(2.0*N);
			c[1]=c[1]+ADD; ADD=ADD/(2.0*N+1.0);
			c[2]=c[2]+ADD; ADD=-E2*ADD; N=N+1.0;
		}
		while (Math.abs(ADD)>= EPS);
	}

	public static void ParabelKoord (double JulDat, double PerDat, 
			double PerDist, 
			double Exz, double Neig, double AKn, double PerL,
			double [] lbr )  //Rad, hzL, hzB: hpNumTyp);
	/*wie HeliozKoord. fuer Parabeln*/
	/*nach MontPfle S. 69,  alle Winkel in Bogenmass */

	/*---------------------------------------------------------------------------*/
	/* PARAB: Berechnung des Orts- und Geschwindigkeitsvektors                   */
	/*        fuer parabolische und nahe-parabolische Bahnen nach Stumpff        */
	/*                                                                           */
	/*        T0   Zeitpunkt des Periheldurchgangs      X,Y    Ortsvektor        */
	/*        T    Berechnungszeitpunkt                 VX,VY  Geschwindigkeit   */
	/*        Q    Periheldistanz                                                */
	/*        ECC  Exzentrizitaet                                                */
	/*        (T0,T in julianischen Jahrhunderten seit J2000)                    */
	/*---------------------------------------------------------------------------*/
	/*public void PARAB(T0,T,Q,ECC:REAL;double X,Y,VX,VY:REAL);*/
	{

		final double EPS    = 1e-9;
		final double KGAUSS = 0.01720209895;
		final int    MAXIT  = 20;
		double E2,E20,FAC, TAU,A,U,U2;
		int I;
		double [] c  = new double [3];

		/*    writeln ('Parabel');*/
		E2=0.0;  FAC=0.5*Exz;  I=0;
		/*    K   = KGAUSS / Math.sqrt(PerDist*(1.0+Exz));*/
		TAU = KGAUSS *(JulDat-PerDat);
		do {
			I=I+1;
			E20=E2;
			A=1.5*Math.sqrt(FAC/(PerDist*PerDist*PerDist))*TAU;
			A=Math.cbrt(Math.sqrt(A*A+1.0)+A);
			U=A-1.0/A;  U2=U*U;  E2=U2*(1.0-Exz)/FAC;

			stumpff (E2,c); 
			FAC=3.0*Exz*c[2];
		}
		while ((Math.abs(E2-E20)>= EPS) || (I<=MAXIT));


		if (I==MAXIT) {Log.d (TAG, " Iteration convergence problem in EphemUtils.PARABELKOORD");}

		double [] xy = new double [2];
		double [] WaAnomRad = new double [2];
		double [] xyz = new double [3];
		
		// x, y Koordinaten in Bahnebene
		xy [0] =PerDist*(1.0-U2*c[1]/FAC);
		xy [1] =PerDist*Math.sqrt((1.0+Exz)/FAC)*U*c[0];

		AstroUtils.cartesianToPolar2(xy, WaAnomRad);
		/*    writeln (raddeg(waanom):10:5, rad:10:5);*/
		WaAnomRad [0] += PerL; 
		AstroUtils.polarToCartesian2(WaAnomRad, xy); //   (x, y, Rad, ((PerL) +waAnom) ); /*Koordinaten in Bahnebene*/
		/*^Hier kein AKn abziehen!!!!*/
		xyz[0] = xy[0];
		xyz[1] = xy[1]*Math.cos(Neig);
		xyz[2] = xy[1]*Math.sin(Neig);
		
		AstroUtils.cartesianToPolar3 (xyz, lbr);
		lbr[0] += +AKn; /*Laenge*/
	} /*of ParabelKoord ***************************************************/



	/*---------------------------------------------------------------------------*/
	/* HYPANOM: Berechnung der exzentrischen Anomalie H=HYPANOM(MH,ECC) zur      */
	/*          mittleren Anomalie MH und zur Exzentrizitaet ECC bei             */
	/*          hyperbolischen Bahnen     MontPfleg S. 65                        */               
	/*---------------------------------------------------------------------------*/

	public static double HYPANOM (double MH, double Exz) {

		final double EPS   = 1e-10;
		final int    MAXIT = 15;
		double H, F, EXH, SINHH, COSHH;
		int    I;

		H=Math.log(2.0 *Math.abs(MH) /Exz+1.8); 
		if (MH<0.0) {H=-H;}
		EXH   =Math.exp(H);
		SINHH =0.5*(EXH-1.0/EXH); COSHH=0.5*(EXH+1.0/EXH);
		F = Exz*SINHH-H-MH; I=0;
		while ( (Math.abs(F)>EPS*(1.0+Math.abs(H+MH))) && (I<MAXIT) ) 
		{
			H = H - F / (Exz*COSHH-1.0);
			EXH =Math.exp(H);
			SINHH=0.5*(EXH-1.0/EXH); COSHH=0.5*(EXH+1.0/EXH);
			F = Exz*SINHH-H-MH; I=I+1;
		}
		if (I>=MAXIT) {Log.d(TAG, " Iteration convergence problem in EphemUtils.HYPANOM");}
		return H;
	}  /*HYPANOM*/




	public static void HyperbelKoord (double JulDat, double PerDat, double Halba, double Exz, 
			double Neig, double AKn, double PerL,
			double [] lbr) //Rad, hzL, hzB: hpNumTyp);
	/*wie HeliozKoord. fuer Hyperbeln*/
	/*nach MontPfle S. 65,  alle Winkel in Bogenmass */

	/*---------------------------------------------------------------------------*/
	/* HYPERB: Berechnung des Orts- und Geschwindigkeitsvektors                  */
	/*         fuer hyperbolische Bahnen     (MontPfle S. 66)                                    */
	/*                                                                           */
	/*        T0   Zeitpunkt des Periheldurchgangs         X,Y   Ortsvektor      */
	/*        T    Berechnungszeitpunkt                    VX,VY Geschwindigkeit */
	/*        Halba    grosse Halbachse (Vorzeichen beliebig)                        */
	/*        Exz  Exzentrizitaet                                                */
	/*        (T0,T in julianischen Jahrhunderten seit J2000)                    */
	/*---------------------------------------------------------------------------*/
	/*public void HYPERB(T0,T,Halba,Exz:hpNumTyp;double X,Y,VX,VY:hpNumTyp);*/
	{
		final double KGAUSS = 0.01720209895;
		double K, MH, H, EXH, COSHH, SINHH, FAC;

		/*    writeln ('Hyperbel');*/
		Halba   = Math.abs(Halba);
		K   = KGAUSS / Math.sqrt(Halba);
		MH  = K*(JulDat-PerDat)/Halba;
		H   = HYPANOM (MH, Exz);
		EXH = Math.exp(H);
		COSHH=0.5*(EXH+1.0/EXH);  SINHH=0.5*(EXH-1.0/EXH);
		FAC = Math.sqrt((Exz+1.0)*(Exz-1.0));   /*RHO = Exz*COSHH-1.0;*/

		double [] xy = new double [2];
		double [] WaAnomRad = new double [2];
		double [] xyz = new double [3];
		
		// x, y Koordinaten in Bahnebene
		xy [0] =Halba*(Exz-COSHH);
		xy [1] =Halba*FAC*SINHH;

		AstroUtils.cartesianToPolar2(xy, WaAnomRad);
		WaAnomRad [0] += PerL; 
		AstroUtils.polarToCartesian2(WaAnomRad, xy); //   (x, y, Rad, ((PerL) +waAnom) ); /*Koordinaten in Bahnebene*/
		/*^Hier kein AKn abziehen!!!!*/
		xyz[0] = xy[0];
		xyz[1] = xy[1]*Math.cos(Neig);
		xyz[2] = xy[1]*Math.sin(Neig);
		
		AstroUtils.cartesianToPolar3 (xyz, lbr);
		lbr[0] += AKn; /*Laenge*/
	}    /*HyperbelKoord*/


	/** Heliocentric coordinates for elliptic orbits
	 * 
	 * @param miAnom
	 * @param Halbachs
	 * @param exz
	 * @param Neig
	 * @param AKn
	 * @param PerL (was PerL +AKn), AKn removed
	 * @param lbr
	 */
	public static void EllipticKoord (
			double miAnom, double Halbachs, double exz, 
			double Neig, double AKn, double PerL, // AKn not added!
			double [] lbr)
	// after Montenbruck p. 63 f,  all in radians
	{
		double  exAnom, exAnomAlt;
//		Log.d("PlanetData", String.format("EphemUtils.EllipticKoord miAnom = %8.5f rad,  exz = %8.5f, PerL = %8.5f, Halbachs = %8.5f AE",
//				miAnom, exz, PerL, Halbachs));

		miAnom = AstroUtils.realMod (miAnom, 2.0*Math.PI);
		exAnom = miAnom;  //Startvalue for Kepler's equation
		do {
			exAnomAlt = exAnom;
			exAnom  = miAnom +exz*Math.sin(exAnom);
		}  // Wepner p. 191
		while (Math.abs(exAnom-exAnomAlt)> 0.0000001);
		exAnom = AstroUtils.realMod (exAnom, Math.PI*2.0);


		double [] WaAnomRad = new double [2];
		double [] xy = new double [2];
		WaAnomRad[1] = Halbachs *(1.0 -exz *Math.cos(exAnom));  // radius
		WaAnomRad[0] = 2.0 * Math.atan( Math.sqrt((1+exz)/(1.0-exz) )  *Math.tan(exAnom/2.0) );

//		Log.d("PlanetData", String.format("EphemUtils.EllipticKoord exAnom = %8.5f rad,  WaAnomRad = %8.5f, r = %8.5f AE",
//				exAnom, WaAnomRad[0], WaAnomRad[1]));

		WaAnomRad[0] += PerL;
		AstroUtils.polarToCartesian2(WaAnomRad, xy); //Koordinates in orbital plane
		
		double [] xyz = new double [3];
		xyz[0] = xy[0];
		xyz[1] = xy[1]*Math.cos(Neig);
		xyz[2] = xy[1]*Math.sin(Neig);
		
		AstroUtils.cartesianToPolar3 (xyz, lbr);
		lbr[0] += AKn; //Length
//		Log.d("PlanetData", String.format("EphemUtils.EllipticKoord hzL = %8.4f°,  hzB = %8.4f°", Math.toDegrees(lbr[0]), Math.toDegrees(lbr[1])));
	} // of EllipticKoord ***************************************************


	/** heliocentric cartesian coordinates for asteroids and comets
	 * 
	 * @param julDat TDT of calculation. lightTimeCorrection subtracted
	 * @param oe     OrbElem orbital elements
	 * @param lbr    double [3] returns heliocentric polar coordinates
	 */
	public static void AsteroidLBR (double julDat, 
			OrbElem oe,
			double [] lbr)
	{
		final double MNull=Math.toRadians(5.0); //Montpfleg. S. 74 (Kepler)
		final double EPS = 0.1;
		double Halba = oe.getSemimajorAxis();
		double invAchs;
		double DELTA  = Math.abs(1.0 -oe.eccentricity);
		double perDat = oe.perihelionTime;
		if (perDat == 0.0) perDat = oe.calcPerihelionJD(); // if not initialized
//		oe.argumentOfPerihelion = 7.65349;   Log.d("PlanetData", "EphemUtils.AsteroidLBR faked Data"); // TODO
		if (Halba == 0.0) {
			invAchs = DELTA / oe.perihelionDistance;  //= 1/a
			Halba = 1.0 / invAchs;
		} else invAchs = 1.0 / Halba;
		
		double miAnom, tau;

		if ((oe.meanAnomaly != 0.0) /* && (TM0 != 0)*/ )  {    // if M0 and TM0 are known (Astroids), then use this value
			tau    = AstroUtils.KGAUSS*(julDat -oe.epochOfOsculation);
			miAnom = oe.meanAnomaly +tau/ Math.sqrt (Halba*Halba*Halba);
		}
		else {
			tau    = AstroUtils.KGAUSS *(julDat -perDat);
			miAnom = tau / Math.sqrt (Halba*Halba*Halba);
//			Log.d("PlanetData", String.format("EphemUtils.AsteroidLBR julDat-LLz = %15.6f, PerDat=  %15.6f, tau = %9.6f,   miAnom = %8.5f", julDat, perDat, tau, miAnom));
		}

		double lbr0 [] = new double [3];
		if (((miAnom < MNull) && (DELTA < EPS)) || (oe.eccentricity == 1.0)) {
			ParabelKoord (julDat, perDat, oe.perihelionDistance, oe.eccentricity, 
					oe.inclination, oe.longitudeOfAscendingNode, oe.argumentOfPerihelion, 
					lbr0);
		}
		else if  (oe.eccentricity < 1.0) {
			//Ellipse
			EllipticKoord (miAnom, Halba, oe.eccentricity,                         
					oe.inclination, oe.longitudeOfAscendingNode, oe.argumentOfPerihelion,
					lbr0);
		} else //Hyperbel
			HyperbelKoord (julDat, perDat, 1.0 / invAchs, oe.eccentricity,
					oe.inclination, oe.longitudeOfAscendingNode, oe.argumentOfPerihelion, 
					lbr0);
//		Log.d("PlanetData", String.format("EphemUtils.AsteroidLBR preRed  hzL = %8.4f°,  hzB = %8.4f°", Math.toDegrees(lbr0[0]), Math.toDegrees(lbr0[1])));
//		Log.d (TAG, "EphemUtils.AsteroidLBR, PlanetData elements Equinox: " + DateTimeUtils.YMDStrg (oe.equinoxOfElements, 0) +String.format("%15.5f", oe.equinoxOfElements));

		AstroUtils.redEcl(lbr0, lbr, oe.equinoxOfElements, julDat);
//		Log.d("PlanetData", String.format("EphemUtils.AsteroidLBR postRed hzL = %8.4f°,  hzB = %8.4f°", Math.toDegrees(lbr[0]), Math.toDegrees(lbr[1])));
	} //AsteroidLBR, return to PlanetoidData.planetoidCalc

/*

  Bahndatyp = Record
                  Name: String [44];
                  //Groesse, Aenderg / Jahrhundert :
                  Halba,
                  Exz,  dExz,
                  Neig, dNeig,
                  AKn,  dAKn,
                  PerL, dPerL,
                  M0,   dM: real;
                  TM0:  JDTyp;          // Zeitpunkt, fuer den M0 gilt
                end;

  CometBahnTyp  = record
                  BahnDat:      Bahndatyp;
                  PerDist:      hpNumTyp;
                  PerDat:       JDTyp;
                  ElemEpoch,
                  absMag, dMag: lpnumTyp;
                  Asteroid:     boolean;   //true b. Aster. wg. Helligeitsformel
                  end;



		public void CometCalc (CometBahn: CometBahntyp;  JulDat: JDTyp;
		double eklL, eklB, Entf, rPlan, Elong: hpNumTyp;
		double mag: lpNumTyp;
		double RSun: hpNumTyp;
		LightTime: Boolean);
		//aequinoktium der Bahnelemente!  Auch fuer Asteroiden geeignet

		const      //Montpfleg. S. 74 (Kepler)
		MNull=5.0*RadFakt; EPS=0.1;

		double  Lichtlaufzeit: double;
		DELTA, InvAchs, TAU, PhasW,
		//exz0, Neig0, AKn0, Perl0,
		miAnom, LPlan, BPlan,
		dummy1, dummy2, dummy3,
		LSun, BSun, xSun, ySun, zSun, x, y, z:    hpNumTyp;
		n,  MaxN: integer;

		{
			with CometBahn do with Bahndat do {

				SonnecalcVSOP (JulDat, LSun, BSun, RSun, false, false);   // Sonnenkoordinaten Aktuell
				RedEkl (lSun, BSun, JulDat, AeqJulDat(ElemEpoch));        // Sonnenkoordinaten auf Bezugsaeq. reduzieren
				Rechtw3D (xSun, ySun, zSun, RSun, LSun, BSun);            // Kartes. Sonnenkoord.
				//Montpfleg. S. 74 (Kepler)
				DELTA  = ABS(1.0-Exz);
				InvAchs= DELTA / PerDist;  //= 1 /Halbachs

				Entf = 1;  //Startwert fuer Iteration der Lichtlaufzeit
				if  LightTime then MaxN = 2 else MaxN = 1;

				for n = 1 to MaxN do {   // Schleife Berechnung Lichtlaufzeit
					if LightTime then Lichtlaufzeit = Entf *tlight else Lichtlaufzeit = 0;
					if (M0 <> 0) and (TM0 <> 0) then {    // Wenn M0 und TM0 bekannt (Astroiden), dann diesen Wert benutzen
						Tau    = KGAUSS*(JulDat -TM0 -Lichtlaufzeit);
						MiAnom = M0 +TAU/ Math.sqrt (Halba*Halba*Halba);
						end
						else {
							TAU    = KGAUSS*(JulDat -PerDat -Lichtlaufzeit);
							MiAnom = TAU / Math.sqrt (Halba*Halba*Halba);
						}
						//   MiAnom = TAU*Math.sqrt(InvAchs*InvAchs*InvAchs); //Bogenmass

						IF ( (MiAnom < MNull) AND (DELTA < EPS) ) or (Exz = 1.0)
						THEN  ParabelKoord (JulDat-Lichtlaufzeit, PerDat, PerDist, Exz,       // Getestet: OK
								Neig, Akn, PerL,  Rplan, LPlan, BPlan)
						ELSE IF (Exz < 1.0)
						then //Ellipse
						HeliozKoord(miAnom, Halba, Exz,                                // Astrweph.inc
								Neig, Akn, PerL + AKn,  Rplan, LPlan, BPlan)
						else //Hyperbel
							HyperbelKoord (JulDat-Lichtlaufzeit, PerDat, 1.0 / InvAchs Exz,
									Neig, Akn, PerL,  Rplan, LPlan, BPlan);

					//  ShowMessage ('Komet HeliozK: ' +Genstring(raddeg (LPlan), 4, 1) + Genstring(raddeg (BPlan), 10, 1) + ' r= '+Genstring (rPlan, 1, 3));

					Rechtw3D (x, y, z, RPlan, LPlan, BPlan);                  // Heliozentrische Kartesische Koordinaten
					Polar3D (xSun+x, ySun+y, zSun+z, Entf, eklL, eklB);       // Geozentr. Koord.
					} // for -Schleife

					Elong = arccos(Math.cos(eklB) *Math.cos(eklL -LSun));
					if Math.sin(eklL -LSun) < 0
					then Elong = -Elong;

					//Helligkeit berechnen
					if Asteroid
					then {
						//Fuer Planetoidenhelligkeit
						PhasW = ArcCos ((sqr(entf) +sqr(rPlan) -sqr(rSun))
								/ (2 * Entf * rPlan) ); //mont 111
						//if Elong < 0 then PhasW = -PhasW;
						//Wepner 235 gibt 0.023 an. in File Mittel 0.15
						mag = AbsMag + dmag * abs (Raddeg (PhasW)) *0.023/0.15
						+ 5.0 * ln (rPlan * Entf) / ln10;
						end
						else
							mag = AbsMag + 5.0 * ln (Entf) / ln10
							+ 2.5 * dmag * ln (rPlan)/ ln10;
					} //With CometBahn
				} //of Cometcalc ***************************************************************
*/
	
	//---------------------------------------------------------------------------
	// NUTEQU: Transformation von mittleren in wahre aequatoriale                
	//	         Koordinaten (Terme >0.1" nach IAU 1980)                           
	//	         Nach Montenbruck NUTEQU                                           
	//---------------------------------------------------------------------------
	
	/** Nutation transformation for mean to true equatorial coordinates
	 * 
	 * @param julDat Julian date
	 * @param args RA, DE in radians, (Radius)
	 */
	public static void Nutat_Aeq (double julDat, double args[]) {

		final double ARC =206264.8062;        // Bogensekunden pro radian = 3600*180/pi 
		double   LS,D,F,N,EPS ;
		double   DPSI,DEPS,C,S;
		double   DX,DY,DZ     ;
		double   T            ;

		T   = (julDat-2451545.0)/36525.0;
		
		double xyzVec[] = new double[3];
		MathOps.polarToCartesian (xyzVec, args[0], args[1], args [2]);
		LS  = TWO_PI * AstroUtils.frac(0.993133+  99.997306*T);   // mittl. Anomalie Sonne    
		D   = TWO_PI * AstroUtils.frac(0.827362+1236.853087*T);   // Diff. Length Moon-Sun
		F   = TWO_PI * AstroUtils.frac(0.259089+1342.227826*T);   // Knotenabstand            
		N   = TWO_PI * AstroUtils.frac(0.347346-   5.372447*T);   // Laenge des aufst.Knotens 
		EPS = 0.4090928-2.2696E-4*T;             	   // Ekliptikschiefe          
		DPSI = ( -17.200*Math.sin(N)   - 1.319*Math.sin(2*(F-D+N)) - 0.227*Math.sin(2*(F+N))
				 + 0.206*Math.sin(2*N) + 0.143*Math.sin(LS) ) / ARC;
		DEPS = ( + 9.203*Math.cos(N)   + 0.574*Math.cos(2*(F-D+N)) + 0.098*Math.cos(2*(F+N))
				 - 0.090*Math.cos(2*N)                 ) / ARC;
		C = DPSI*Math.cos(EPS);  S = DPSI*Math.sin(EPS);
		DX = -(C*xyzVec[1]+S*xyzVec[1]); DY = (C*xyzVec[0]-DEPS*xyzVec[1]); DZ = (S*xyzVec[0]+DEPS*xyzVec[1]);
		xyzVec[0]  =  xyzVec[0] + DX;        xyzVec[1]  = xyzVec[1] + DY;       xyzVec[2]  = xyzVec[2] + DZ;
		MathOps.cartesianToPolar (xyzVec, args);
	}

	/** annual stellar aberration according Meeus P. 139
	 *
 	 * @param julDat
	 * @param ekl_LB_rad, longitude and latitude, returns result here
	 */
	public static void Aberration (double julDat, double ekl_LB_rad []) {

		// Meeus p. 139
		final double k = Math.toRadians (20.49552/3600.0); //constant of aberration (20.49552" used; Meeus 1991,p 139)
		double T =  (julDat-2451545.0)/36525.0;
		double solarLon = calcSunLon (julDat); // geometric solar longitude
		double e = 0.016708617 - 0.000042037 *T - 0.0000001236 *T*T; //eccentricity of the Earth's orbit:
		double pi_rad = Math.toRadians (102.93735 +0.71953 *T +0.00046*T*T); //longitude of the perihelion of the Earth's orbit

		double d_ekl_L = (-k *Math.cos (solarLon -ekl_LB_rad [0])
				+ e*k*Math.cos (pi_rad -ekl_LB_rad [0]))
				/ Math.cos (ekl_LB_rad [1]);
		double d_ekl_B = -k *Math.sin (ekl_LB_rad [1]) *
				(Math.sin (solarLon -ekl_LB_rad [0]) - e*Math.sin (pi_rad -ekl_LB_rad [0]));

//		Log.d ("nutation aberration", String.format("d_ekl_L: %.2f sec,  d_ekl_B: %.2f sec", Math.toDegrees(d_ekl_L)*3600.0,  Math.toDegrees(d_ekl_B)*3600.0));
		ekl_LB_rad [0] -= d_ekl_L;
		ekl_LB_rad [1] -= d_ekl_B;
	}
	
	/** gets angular separation (elongation) in radians from the sun
	 * 
	 * @param eklL ecliptic longitude of the planet
	 * @param eklB ecliptic longitude of the planet
	 * @param lSun ecliptic longitude of the sun
	 * @return angular elongation in radians
	 */
	public static double Elongation (double eklL, double eklB, double lSun) {
		double elong = Math.acos(Math.cos(eklB) *Math.cos(eklL -lSun));
	    if (Math.sin(eklL -lSun) < 0) elong = -elong;
	    return elong;
	}
	
	/** gets angular separation (elongation) of planet in radians from the sun
	 * 
	 * @param pD PlanetData instance
	 * @return angular elongation in radians
	 * @throws NoInitException 
	 */
	public static double Elongation (PlanetData pD) {
		
		try {
			return Elongation (pD.getEclipticLon(), pD.getEclipticLat(), pD.getSolarLon());
		} catch (NoInitException e) {
			//e.printStackTrace();
			Log.d("Elongation:", "NoInitException!");
			return 0.0;
		}
	}

	
	public static void TopozReduktion (double args [], // Ra, De, Entf // Bogenmass, AE
    		double julDat,
    		double lon, double lat)  // degrees!
		//Reduces Koordinates do Topocentric
	{
		//double x, y, z, x2, y2, z2;
	    double tmpVec1[] = new double[3];
	    double tmpVec2[] = new double[3];
	    MathOps.polarToCartesian (tmpVec1, args);
	    MathOps.polarToCartesian (tmpVec2,
	    		Math.toRadians(AstroUtils.sternzeit(julDat, lon)*15.0), Math.toRadians(lat), AstroUtils.R_EARTH/AstroUtils.AE);  
	    tmpVec1 [0] -= tmpVec2 [0];
	    tmpVec1 [1] -= tmpVec2 [1];
	    tmpVec1 [2] -= tmpVec2 [2];
	    AstroUtils.cartesianToPolar3  (tmpVec1, args);
	} //TopozReduktion **************************************
	


	
	// Iteration fuer Mondphasen
	private final static double PHASE_DUR = LunarCalc.SYNODIC_MONTH/4.0;
	
	
	/**  calculates next moon phase (new, Quarters, Full) closest to jD 
	 * 
	 * @param jD julian date
	 * @return next moon phase (new, Quarters, Full) closest to jD 
	 */
	private static double IteratePhase (double jD) {
		double ageDiff, iterationResult;
		int   i = 0;
		ageDiff = 0.0;
		iterationResult = jD;
		do {
			ageDiff= AstroUtils.realMod (LunarCalc.ageOfMoonInDays(iterationResult) +3, PHASE_DUR) -3;
			iterationResult = iterationResult-ageDiff;
			i++;
		} while ((Math.abs(ageDiff) > (1.0/4000.0)) && (i < 20)); //(Fehler unter 1 Min) , max 10 Iterationen
//		Log.d("Iteration", ""+i);
		return iterationResult;
	}
	
	/** Calculates last and next three moon phases from JulDat
	 * @param  julDat
	 * @param  args [4] returns the phases in JulDat
	 * @return Phase of Arg [0] (0 = New Moon, 1 =1st Quarter, 2= Full.... 
	 */
	public static int MoonPhases (double julDat, double args[]) {
		
		double age = LunarCalc.ageOfMoonInDays(julDat);
		int phase0 = (int) Math.floor (age/PHASE_DUR);

		args [0] = IteratePhase (julDat-AstroUtils.realMod (age, PHASE_DUR));
		args [1] = IteratePhase (args [0] +PHASE_DUR);
		args [2] = IteratePhase (args [1] +PHASE_DUR);
		args [3] = IteratePhase (args [2] +PHASE_DUR);
		return phase0;
	}
	
	/**  calculates the last and next  new moon from JulDat
	 * 
	 * @param julDat
	 * @param args [2] returns the Last [0] and next [1] new Moon in JulDat
	 * @return
	 */
	public static void NewMoons (double julDat, double args[]) {
		double age = LunarCalc.ageOfMoonInDays(julDat);
		args [0] = IteratePhase (julDat-age);
		args [1] = IteratePhase (args [0] +LunarCalc.SYNODIC_MONTH);
	}
	
	/**
	 * Constant for iteration for AspectFinder. Average sidereal velocity per day
	 */
	private static double [] PLANET_RADS_PER_DAY =
	{	Math.PI /( 1.0 * 182.622),        // Sun = Earth
		Math.PI /( 0.248085 * 182.622),   // Mercury
		Math.PI /( 0.61521  * 182.622),   // Venus
		Math.PI /( 1.0      * 182.622),   // Earth
		Math.PI /( 1.88089  * 182.622),   // Mars
		Math.PI /( 11.862   * 182.622),   // Jupiter
		Math.PI /( 29.458   * 182.622),   // Saturn
		Math.PI /( 86.01    * 182.622),   // Uranus
		Math.PI /( 164.8    * 182.622),   // Neptune
		Math.PI /( 297.7    * 182.622),   // Puto
		Math.PI /( 27.322),               // Moon
		Math.PI /( 1.0      * 182.622),   // Planetoid, set to Earth
	};
		
	/**
	 * 
	 * @param planet   Planets.SUN .. Planets.LUNA
	 * @param tdt      Dynamic Time in Julian Date format
	 * @param forward  Calculate into future, else backward
	 * @param opposition (inferior conjunction) if true, superior conjunction if false 
	 * @return julDat of Aspect, if error return -1.0
	 */
	public static double AspectFinder (int planet, OrbElem orbElem, double tdt, boolean forward, boolean opposition) {
	
		final int MAX_ITERATION = 100;

		if (planet == Planets.EARTH || planet == Planets.SUN) return tdt;     // not defined for Earth and Sun  

     	final double ITERATION_LIMIT = 1./(24. *60.);    // Iteration break condition in days (1 Minute)
     	double iterationFactor = 1.;              //
     	//double  direction      = 1.;               // may be 1 or -1
		final double ITERATION_TOLERANCE = AstroUtils.degRad(20.);  // Tolerance for switching to next  
	    final ObsInfo oi 	= new ObsInfo ();  //dummy variable
	    double dT;
	    double lastdT		= 1.;
	    double lastdL	    = 999;
	    int iterationCount 	= 0;
	    boolean reverse    	= false;
	    
		Log.d("Aspect Iteration", DateTimeUtils.YMDStrg(tdt, 0) +"  Start, Opposition: " + opposition  +"   forw: "+ forward); 
		do {// Calculate heliocentric longitude of Earth and Planet
			PlanetData pdE = new PlanetData ();
			double lonE	   = pdE.calcLon(Planets.EARTH, tdt, oi); // calcs length of Sun
			PlanetoidData pdP = new PlanetoidData(orbElem);
			double lonP = pdP.calcLon(planet, tdt, oi);
			if (opposition) lonE -= Math.PI; // for opposition set LonEarth -= 180 deg;
			
			if (forward) {while (lonE > lonP +ITERATION_TOLERANCE) lonE -= TWO_PI;} // set LonE < LonP for forward
			else         {while (lonE < lonP -ITERATION_TOLERANCE) lonE += TWO_PI;} // set LonE > LonP for backward

			//Norm elongation to -TWO_PI .. TWO_PI
			double deltaLon = AstroUtils.latAngle(lonE-lonP);		
			
			if (planet != Planets.PLANETOID || Math.abs(orbElem.eccentricity) < 0.9) {
				// iterate times
				dT = deltaLon / (PLANET_RADS_PER_DAY [planet] - PLANET_RADS_PER_DAY [Planets.EARTH]);
				
				if (lastdT != 0 && lastdT !=dT) {iterationFactor = lastdT /(lastdT -dT);} 
				  else  {iterationFactor = 1.0;} 
				if (iterationFactor > 2 || iterationFactor < 0.5) {iterationFactor = 1.0;} 
			}
		
			else { // Kometen und Obj. mit hoher Exz.
				double ddL = AstroUtils.latAngle(deltaLon - lastdL); // aenderung von deltaLon

				if (iterationCount > 1) {// Vermeiden, ueber den Zielbereich hinauszuschiessen
					if  (Math.abs(ddL)*AstroUtils.DEGFAKT > Math.abs(lastdT)){
						iterationFactor *= 0.5; // Vermeiden, ueber den Zielbereich hinauszuschiessen
//						Log.d("Aspect Iteration", "  ddL: " +AstroUtils.DMSStrg (ddL, "d3"));
					}
				}

				if (Math.abs(deltaLon) < ITERATION_TOLERANCE){	// nur im Zielbereich Richtung aendern!					
					if (Math.abs(deltaLon) > Math.abs(lastdL)) {   // ? Approximation wird schlechter?						 
						iterationFactor *= -0.5;
					} 
					else if (Math.abs(ddL)*AstroUtils.DEGFAKT < Math.abs(lastdT)/2
							&& iterationCount > 1) 
						iterationFactor *= 1.2; // ggf vorsichtig erhoehen
				}
				lastdL = deltaLon;


				// iterate times
				dT = deltaLon / (PLANET_RADS_PER_DAY [Planets.MARS]); 
				// Limit dT
				if (dT >  30) dT =  30;
				if (dT < -30) dT = -30;
				if (Math.abs(deltaLon) < ITERATION_TOLERANCE) dT /= 2; // Limit dT within +-40 deg

				if (Math.abs(deltaLon) > ITERATION_TOLERANCE) { // Force direction
					if (forward) { dT = Math.abs(dT); } else dT = -Math.abs(dT);
				}
			} // Kometen

			
//			Log.d("Aspect Iteration", "  " + DateTimeUtils.YMDStrg(tdt, 0)
//					+"  dT: " +String.format("%+8.3f",dT)
//					+String.format("   dL: %9s", AstroUtils.DMSStrg (deltaLon, "d3"))
//					+String.format("  IterFact: %6.3f", iterationFactor));

			tdt += dT * iterationFactor;
			lastdT = dT;
			iterationCount ++;
		} while (Math.abs (dT) > ITERATION_LIMIT && iterationCount < MAX_ITERATION);
		Log.d("Aspect Iteration", "Finish, result valid: " +(Math.abs (dT) <= ITERATION_LIMIT)); 
		if (Math.abs (dT) <= ITERATION_LIMIT ) {return tdt;}
		else return -1.0;
	}
	
	/** Returns 1 if next aspect is an opposition, else 0
	 * 
	 * @param planet
	 * @param orbElem not null only for planetoids!
	 * @param tdt
	 * @param tdtTable Table [3] containing last and next two aspects
	 * @return   1 if aspect (tdtTable [0]) is an opposition, else 0, error: -1
	 */
	public static int AspectTable (int planet, OrbElem orbElem, double tdt, double [] tdtTable)
	{   
		if (planet == Planets.PLANETOID) {
			if (orbElem.eccentricity < 0.99){
				double a = orbElem.getSemimajorAxis();
				PLANET_RADS_PER_DAY [planet] = PLANET_RADS_PER_DAY [Planets.EARTH] /Math.sqrt(a*a*a);
			} else PLANET_RADS_PER_DAY [planet] = PLANET_RADS_PER_DAY [Planets.MARS]; // Not earth, else division by zero!
		}
		
		double halfSynodicTime = Math.PI/Math.abs(PLANET_RADS_PER_DAY [planet] - PLANET_RADS_PER_DAY [Planets.EARTH]);
		if (planet == Planets.PLANETOID && orbElem.eccentricity > 0.95) halfSynodicTime = 40; 
		Log.d("Aspect Table", "HalfSynTime: " +halfSynodicTime);

		// Begin with search for next opposition
		Log.d("Aspect Table", "Start #1 " +DateTimeUtils.YMDStrg(tdt, 0) +"  (find next opposition / inf. conj.)");
		tdtTable [1] = AspectFinder (planet, orbElem, tdt, true, true);
		if (tdtTable [1] > 0) { // valid value?
			if (tdtTable [1] <= tdt) // ? Opposition was before TDT
			{  	Log.d("Aspect Table", "Find #1   " +DateTimeUtils.YMDStrg(tdt, 0) +"  (find next opposition)");
				tdtTable [1] = AspectFinder (planet, orbElem, tdtTable [1] +halfSynodicTime*2., true, true);
				}	
			if (tdtTable [1] < 0) return -1; // Return error, if invalid value
			
			// search the conjunction before
			Log.d("Aspect Table", "Find #0 a " +DateTimeUtils.YMDStrg(tdt, 0) +"  (find conjunction before)");
			tdtTable [0] = AspectFinder (planet, orbElem, tdtTable [1] -halfSynodicTime , false, false);
			
			if (tdtTable [0] < 0) return 0; // tdtTable [0]) is a conjunction
			if (tdtTable [0] <= tdt){ // ? before TDT
				// calculate next conjunction 
				Log.d("Aspect Table", "Find #2   " +DateTimeUtils.YMDStrg(tdt, 0)+"  (find conjunction after)");
				tdtTable [2] = AspectFinder (planet, orbElem, tdtTable [1] +halfSynodicTime, true, false);
				return 0; // aspect (tdtTable [0]) is an opposition
			}
			else {      //tdtTable [0] is next conjunction, shift up in Table
				tdtTable [2] = tdtTable [1]; // next opposition
				tdtTable [1] = tdtTable [0]; // next conjunction
				// calculate previous opposition
				Log.d("Aspect Table", "Find #0 b " +DateTimeUtils.YMDStrg(tdt, 0)+"  (find opposition before)");
				tdtTable [0] =AspectFinder (planet, orbElem, tdtTable [1] -halfSynodicTime, false, true);
				return 1;			
			}
			
		} else { // tdtTable [1] is invalid value
			tdtTable [0] = -1;
			tdtTable [1] = -1;
			tdtTable [2] = -1;
			Log.d("Aspect Table", "Invalid values");
			return -1;
		}
	}
	
	
	/** Calculates time of maximum elongation
	 * 
	 * @param planet
	 * @param tdt
	 * @param args 0: tdt of max. elongation, 1: max elongation in radians
	 */
	public static void MaxElongationFinder (int planet, double tdt, double args[])
	{   if (planet != Planets.MERCURY && planet != Planets.VENUS) {
		args [0] = 0.;
		args [1] = 0.;
		// only defined for inner planets  
	}
	
		final double ITERATION_LIMIT = 1./50.;    // Iteration break condition in days
		final double ITERATION_TOLERANCE = AstroUtils.degRad(20.);  // Tolerance for switching to next  
	    final ObsInfo oi = new ObsInfo ();  //dummy variable
	    double dT, elong, lastElong;
//		Log.d("Elongation Iteration: ", DateTimeUtils.YMDStrg(tdt, 0) +"  Start"); 
		
		// Calculate heliocentric longitude of Earth and Planet
		PlanetData pdP = new PlanetData (planet, tdt, oi);
		double lonE = 0; 
		double lonP = 0;
		try {
			lonE = pdP.getSolarLon() +Math.PI;
			lonP = pdP.getPolarLon();
		} catch (NoInitException e1) {
			args [0] = 0.;
			args [1] = 0.;
			Log.d("MaxElongationFinder", "NoInitException!");
		}
//		PlanetData pdP = new PlanetData();
//		double lonP = pdP.calcLon(planet, tdt, oi);

		//Norm elongation to -PI .. PI
		double deltaLon = lonE-lonP;
		while (deltaLon >  Math.PI) deltaLon -= TWO_PI; 
		while (deltaLon < -Math.PI) deltaLon += TWO_PI;

		
		// Start with deltaLon = 90 deg
		if (deltaLon >= 0) {deltaLon -= HALF_PI;} else {deltaLon += HALF_PI;} 
		dT = deltaLon / (PLANET_RADS_PER_DAY [planet] - PLANET_RADS_PER_DAY [Planets.EARTH]);
//		Log.d("Elongation Iteration: ", "  " +DateTimeUtils.YMDStrg(tdt, 0) +"  dT: " +String.format("%+.3f",dT) +"   dLong: " +DMSStrg (deltaLon, "d3")); 		tdt += dT;
		
		dT = 5.0;
		new PlanetData (planet, tdt, oi);
		lastElong = Elongation (pdP);
		do {
			pdP = new PlanetData (planet, tdt+dT, oi);
			elong = Elongation (pdP);
			if (Math.abs(elong) < Math.abs(lastElong)){
				dT *= -0.35;  //Change iteration direction, smaller value
			}
			else {
				tdt +=dT; 
				lastElong = elong;
			}
//			Log.d("Elongation Iteration: ", "  " +DateTimeUtils.YMDStrg(tdt, 0) +"  dT: " +String.format("%+.3f",dT) +"   Elong: " +DMSStrg (elong, "d3")); 
		} while (Math.abs (dT) > ITERATION_LIMIT);
//		Log.d("Elongation Iteration: ", "Finish!"); 
		args[0] = tdt;
		args[1] = elong;
	}

	/** calculates longitude of Sun at tdt JulDat Time
	 *
	 * @param tdt
	 * @return
	 */
	public static double calcSunLon (double tdt) {
		final ObsInfo oi = new ObsInfo ();  //dummy variable
		double dT;
//		Log.d("sunLonFinder", "  " +DateTimeUtils.YMDStrg(tdt, 0) +"  Start");
		PlanetData pdE = new PlanetData ();
		// Calculate heliocentric longitude of Earth
		double lon	   = pdE.calcLon(Planets.EARTH, tdt, oi) -Math.PI;
		while (lon < -Math.PI) lon += TWO_PI;
		return lon;
	}

//	private static double aberration_lon (double tdt, double lon) {
//		final double k = Math.toRadians (20.49552/3600.0); //constant of aberration (20.49552" used; Meeus 1991,p 139)
//		double T =  (tdt-2451545.0)/36525.0;
//		double e = 0.016708617 - 0.000042037 *T - 0.0000001236 *T*T; //eccentricity of the Earth's orbit:
//		double pi_rad = Math.toRadians (102.93735 +0.71953 *T +0.00046*T*T); //longitude of the perihelion of the Earth's orbit
//		double d_ekl_L = (-k *Math.cos (lon) + e*k*Math.cos (pi_rad -lon));
//		lon -= d_ekl_L;
//		return lon;
//	}

	/** Calculates time of certain ecliptic longitude of the sun (eg. for Season calculations)
	 * 
	 * @param tdt start value, dynamic Time in Julian days
	 * @param lon longitude to search for
	 * @return tdt time, when sun has longitude lon
	 */
	public static double sunLonFinder (double tdt, double lon) {
     	final double ITERATION_LIMIT = 1./(24. *1200.0);    // Iteration break condition in days, appr. 5 iterations
	    final ObsInfo oi = new ObsInfo ();  //dummy variable
	    double dT;
		PlanetData pdE = new PlanetData ();

		// apply nutation
		Nutation nutation = new Nutation( AstroOps.toMillenia( tdt ));
		lon -= Math.toRadians(nutation.getDPhi ()/3600.0);

		// apply aberration
		lon += Math.toRadians(21.0/3600.0); // add 21 arc seconds aberration of light

//		Log.d("sunLonFinder", "  " +DateTimeUtils.YMDStrg(tdt, 0) +"  Start" +String.format (" %7.3f", Math.toDegrees(lon)));
		do {// Calculate heliocentric longitude of Earth
			double lonE	   = pdE.calcLon(Planets.EARTH, tdt, oi); 
			
			//Norm elongation to -TWO_PI .. TWO_PI
			double deltaLon = lon-lonE;
			while (deltaLon > Math.PI)  deltaLon -= TWO_PI; 
			while (deltaLon < -Math.PI) deltaLon += TWO_PI;
			
			// iterate times
			dT = deltaLon / PLANET_RADS_PER_DAY [Planets.EARTH];
 			tdt += dT ;
//			Log.d("sunLonFinder", "  " + DateTimeUtils.YMDStrg(tdt, 0) +"  dT: " +String.format("%+.6f",dT) +"   dL: " +AstroUtils.DMSStrg (deltaLon, "d5"));
		} while (Math.abs (dT) > ITERATION_LIMIT);
//		Log.d("sunLonFinder", "SunLonFinder Finish!"); 
		return tdt;
	}
	

	
	
}
