package strickling.utils;

/*****************************************************************************\
 *
 *  Copyright 2019 Dr. Wolfgang Strickling
 *  published under  GNU GPL 3
 *
 *   This file is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 \*****************************************************************************/
 

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;

import android.content.Context;
import android.content.pm.ApplicationInfo;
import android.text.format.DateFormat;
import android.util.Log;

/* Class for using Java and Android formatted outputs
 * 
 * Problem: dates before year 1 will be formatted wrong! (julDat < 1721423)
 * 
 */

// see http://stackoverflow.com/questions/4594519/how-do-i-get-localized-date-pattern-string
// Style can be SHORT, MEDIUM, LONG, FULL, DEFAULT


public class FormatterUtils {

	/* Formats:  
	 *  ((SimpleDateFormat) java.text.DateFormat.getDateInstance(java.text.DateFormat.SHORT,  Locale.getDefault())).toPattern()
	 *      //gives only pattern of the default locale
	 *  
	 *  ((SimpleDateFormat) android.text.format.DateFormat.getDateFormat (this)).toPattern();
	 *      // gives numeric patterns
	 *      // getTimeFormat:   with user's 12-/24-hour clock preference, without seconds!
	 *      
	 *   Settings.System.getString(getContentResolver(), Settings.System.DATE_FORMAT); Settings.System.getString(getContentResolver(), Settings.System.DATE_FORMAT);
	 *   gives full string of user's selection. May contain long date formats    
	 */


	public final static String TAG = "FormatterUtils";
	public final static String RTL = "\u200F"; // Right to left space, for Arabic etc see http://www.w3.org/TR/WCAG20-TECHS/H34.html
	public final static String LTR = "\u200E"; // Left to right space, for English etc
	
	public static final String HPPattern = "yyyy-MM-dd HH:mm:ss.SSSZ";
	public static final String YMDPattern = "yyyy-MM-dd";
	
	public static String replaceArabic (String s) {
		// converting eastern Arabic numbers to western numbers
		return s.replace("\u0660","0")  // caution: There is another Arabic table starting at \u06f0 !
				.replace("\u0661", "1")
				.replace("\u0662", "2")
				.replace("\u0663", "3")
				.replace("\u0664", "4")
				.replace("\u0665", "5")
				.replace("\u0666", "6")
				.replace("\u0667", "7")
				.replace("\u0668", "8")
				.replace("\u0669", "9")
				.replace("\u066B", ".");  // decimal comma
	}
	
	
	public static String replaceFullWidthSpace (String s) {
		// Replace U+0020 (SPACE) by  U+2001	(EM QUAD, full width space) for Chinese langugages
		return s.replaceAll(" ", "\u2001"); 
	}

	/** 
	 * @return "de" or "en_US" 
	 */
	public static String getDefaultLanguage () {
		return getLanguage (Locale.getDefault());
	}
	
	/** 
	 * @return "de" or "en_US" 
	 */
	public static String getLanguage (Locale locale) {
		String s = locale.getLanguage();
		if (! (locale.getCountry().toLowerCase().equals(s) || locale.getCountry().equals("")))  s+="_"+locale.getCountry();
		return s;
	}

	
	/** output user selected date pattern, eg. "yyyy.MM.dd"   force 4 digit year
	 * 
	 * @param context   (usually: this)
	 * @param removeDOW (if true, remove any "E" entries)
	 * @return
	 */
	public static String getDatePattern (Context context, boolean removeDOW){ 
		String datePattern = ((SimpleDateFormat) DateFormat.getDateFormat (context)).toPattern();
		if (! datePattern.contains("yyyy")) datePattern = datePattern.replace("yy", "yyyy");
		if (! datePattern.contains("MM"))   datePattern = datePattern.replace("M", "MM");
		if (! datePattern.contains("dd")) 	datePattern = datePattern.replace("d", "dd");
		if (removeDOW) {
			if (datePattern.contains("E,"))  	datePattern = datePattern.replace("E,", "").replace("E", "").trim(); // Remove DOW
			if (datePattern.contains(", E")) 	datePattern = datePattern.replace(", E", "").replace("E", "").trim();// Remove DOW
		}
		try{
			SimpleDateFormat sdf= new SimpleDateFormat (datePattern);
		} catch (Exception e) {
//			Log.d(TAG, "getTimePattern: invalid timePattern: " +datePattern +" replace by: " +DateTimeUtils.DEFAULT_DATE_FORM);
			datePattern = DateTimeUtils.DEFAULT_TIME_FORM;
		}
		return datePattern;
	}

	// see http://developer.android.com/reference/java/text/SimpleDateFormat.html
	// h: 12 h format without zero, H: 24h with zero (0-23),   k: 24h format without Zero, K: 12 h with zero (0-11)
	/** returns user selected time pattern like "hh:mm:ss", forcing second output, keeping selected 12/24hr fromat
	 * 
	 * @param context
	 * @return
	 */
	public static String getTimePattern (Context context){
		return getTimePattern (context, true);
	}
	
	public static String getTimePattern (Context context, boolean withSeconds){ 
		String timePattern = ((SimpleDateFormat) DateFormat.getTimeFormat (context)).toPattern();
		timePattern = timePattern.replace("k", "H"); // k => SimpleDate Formatter returns  24:00:00h instead of 00:00:00 h!
		if (! timePattern.contains("kk") ) timePattern = timePattern.replace ("k", "kk"); 
		if (  timePattern.startsWith("h") && !timePattern.startsWith("hh")) timePattern = "h"+timePattern;//timePattern.replace("h", "hh");
		if (! timePattern.contains("HH") ) timePattern = timePattern.replace("H", "HH");
		if (! timePattern.contains("ss")) {
			 // some Locales have decimal point separator "hh.mm", Copy character preceding "mm"
			if (timePattern.charAt(timePattern.indexOf("mm")-1) == '\'') {timePattern = timePattern.replace("mm", "mm:ss");} // catch Portugese: "H'h'mm"
			else timePattern = timePattern.replace("mm", "mm" +timePattern.charAt(timePattern.indexOf("mm")-1) +"ss");
		}
		//timePattern = timePattern.replace("h", "H"); // k => SimpleDate Formatter returns  24:00:00h instead of 00:00:00 h!
		if (timePattern.contains("a")) timePattern = timePattern.replace("H", "K").replace("k", "h"); // otherwise error on parsing am/pm times
		
		// check validity of patter, else replace by standard pattern
		try{
			SimpleDateFormat sdf= new SimpleDateFormat (timePattern);
		} catch (Exception e) {
//			Log.d(TAG, "getTimePattern: invalid timePattern: " +timePattern +" replace by: " +DateTimeUtils.DEFAULT_TIME_FORM);
			timePattern = DateTimeUtils.DEFAULT_TIME_FORM;
		}
		if (! withSeconds) timePattern = timePattern.substring(0, timePattern.indexOf("mm")+2);
//		Log.d(TAG, "SimpleDateFormat: " +((SimpleDateFormat) DateFormat.getTimeFormat (context)).toPattern()
//				+"   New pattern: " +timePattern);
		return timePattern;
	}

//	/** returns millis for formatting date and time outputs
//	 * 	
//	 * @param julDat   Julian Date
//	 * @param timeZoneHrs in hours
//	 * @return
//	 */
//	public static long millisForFormatter (double julDat, double timeZoneHrs) {
//		// necessary, as formatter use the system's timeZone for output!
//		return DateTimeUtils.julDat2Millis(julDat +(timeZoneHrs -DateTimeUtils.GetCurrentTimeZone())/24.);
//	}
	
	/** returns String like "GMT+02:00", used by getTimeZoneAbbrRaw*/
	public static TimeZone getTimeZone (double timeZoneHrs) {
		TimeZone tz = TimeZone.getTimeZone("GMT" +DateTimeUtils.MSDurStrg(Locale.US, timeZoneHrs, true, true, 0)); // do  NOT change GMT!!
//		TimeZone tz =  TimeZone.getDefault();
//		tz.setRawOffset((int) (timeZoneHrs*3600000));
//		tz.setRawOffset((int) (timeZoneHrs*3600000)); //Log.d(TAG, "    getTimeZone: " +tz.getRawOffset());
		return tz;
	}

	/** should return timeZone abbreviation like MEZ, MESZ for given JulDat and rawTimeZoneHrs
	 *  does not work, as DST switching values are missing.  Always returns GMT related values like "GMT+02:00"
	 * 
	 * @param locale
	 * @param julDat
	 * @param rawTimeZoneHrs
	 * @return
	 */
    public static String getTimeZoneAbbrRaw (Locale locale, double julDat, double rawTimeZoneHrs) {
		if (locale == null) locale = Locale.US;
    	Calendar cal = Calendar.getInstance(getTimeZone (rawTimeZoneHrs), locale);
    	cal.setTimeInMillis(DateTimeUtils.julDat2Millis(julDat));
    	return String.format(locale, "%tZ", cal);
    }
    
	
	/** returns timeZone abbreviation like MEZ, MESZ for given JulDat and TimeZone
	 * 
	 * @param locale
	 * @param julDat
	 * @param tz
	 * @return
	 */
    public static String getTimeZoneAbbr (Locale locale, double julDat, TimeZone tz) {
		if (locale == null) locale = Locale.US;
    	Calendar cal = Calendar.getInstance(tz, locale);
    	cal.setTimeInMillis(DateTimeUtils.julDat2Millis(julDat));
    	return String.format(locale, "%tZ", cal);
    }
    
	
	
	/** Outputs formatted time string
	 * 
	 * @param julDat
	 * @param timeZoneHrs   default: DateTimeUtils.GetCurrentTimeZone() 
	 * @param pattern       default: FormatterUtils.getTimePattern(this)
	 * @return
	 */
	public static String timeStrg (double julDat, double timeZoneHrs, String pattern) {
		return timeStrg (julDat, timeZoneHrs, pattern, Locale.getDefault());
	}
	
	public static String timeStrg (double julDat, double timeZoneHrs, String pattern, Locale locale) {
//		Log.d("PlanetDroid", "Error: " +pattern);
		if (locale == null) locale = Locale.US;
		SimpleDateFormat sdf= new SimpleDateFormat (pattern, locale);
		sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
		//sdf.setTimeZone(getTimeZone(timeZoneHrs));
		return sdf.format(new Date(
				DateTimeUtils.julDat2Millis(
						AstroUtils.frac(julDat +timeZoneHrs/24.)+2440590. // set JD to 2440590,... to avoid negative year numbers 
						)));
	}

	/** returns date format, for years 0 and before fallback to "yyyy - MM - dd" 
	 * 
	 * @param julDat
	 * @param timeZoneHrs
	 * @param pattern
	 * @return
	 */
	public static String dateStrg (double julDat, double timeZoneHrs, String pattern) {
		return dateStrg (julDat, timeZoneHrs, pattern, Locale.getDefault());
	}

	public static String dateStrg (double julDat, double timeZoneHrs, String pattern, Locale locale) {
		if (locale == null) locale = Locale.US;
		if (julDat < 1721423. || pattern == null || locale == null) { 
			// use astro-routine for dates < 0001-01-01
			return DateTimeUtils.YMDStrg (julDat, timeZoneHrs);
		} 
		else { 
			// use Android-routines for dates >= 0001-01-01
			SimpleDateFormat sdf= new SimpleDateFormat (pattern, locale);
			sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
			//sdf.setTimeZone(getTimeZone(timeZoneHrs));
			return sdf.format(new Date(
					DateTimeUtils.julDat2Millis(julDat +timeZoneHrs/24.)	
//					millisForFormatter(julDat, timeZoneHrs)
			));
		}
	}
	
	public static String dateStrg (long millis, String pattern, Locale locale) {
		if (locale == null) locale = Locale.US;
		SimpleDateFormat sdf= new SimpleDateFormat (pattern, locale);
		sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
		return sdf.format(new Date(millis));
	}

	/** returns String
	 * 
	 * @param tDays, time in fractions of day (0..1)
	 * @param pattern eg hh:mm:ss
	 * @param locale  eg. locale.US
	 * @return
	 */
	public static String formatDays (double tDays, String pattern, Locale locale) {
		if (tDays >= 0 || pattern == null || locale == null) {
			try {
				SimpleDateFormat sdf;
				if (locale == null) locale = Locale.US;
				sdf= new SimpleDateFormat (pattern, locale);
				sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
				return sdf.format(new Date((long) ((tDays)*86400000)));//TODO
			} catch (Exception e) {return "-----";}
		}
		else {
			
		char c;
			try{
				c = pattern.charAt(pattern.indexOf("mm")-1);
			}catch (Exception e) {
					c = ':'; 
			}
			if (c == '\'') c = ':'; // Portugese: "H'h'mm"
			if (pattern.contains("ss")) {return "--" +c +"--" +c +"--";}
			else return "--" +c +"--";
		}
	}


	/** Return formatted Date+Time String, fallback to "yyyy - MM - dd  HH:mm:ss for year 0 and before
	 * 
	 * @param julDat
	 * @param timeZoneHrs
	 * @param pattern 
	 * @return
	 */
	public static String zeitStrg (double julDat, double timeZoneHrs, String pattern) {
		return zeitStrg (julDat, timeZoneHrs, pattern, Locale.getDefault());
	}
	
	public static String zeitStrg (double julDat, double timeZoneHrs, String pattern, Locale locale) {
		julDat += DateTimeUtils.JD_ROUND_PREVENTER;  // Prevent round errors
		if (locale == null) locale = Locale.US;

		if (julDat < 1721423. || pattern == null || locale == null) { 
			// use astro-routines for dates < 0001-01-01
			return DateTimeUtils.zeitStrg (julDat, timeZoneHrs); 
		} 
		else {
			// use Android-routines for dates >= 0001-01-01
			String s = " Error,";
			if (pattern == null || locale == null) {return DateTimeUtils.zeitStrg (julDat, timeZoneHrs);} // fallback to AstroRoutine if Error
			try {
				SimpleDateFormat sdf= new SimpleDateFormat (pattern, locale);
				sdf.setTimeZone(TimeZone.getTimeZone("UTC"));  // ICS does not respect any other time zone!
				//sdf.setTimeZone(getTimeZone(timeZoneHrs)); //Log.d(TAG, "    getTimeZone: " +sdf.getTimeZone().getRawOffset());
				s = sdf.format(new Date(DateTimeUtils.julDat2Millis(julDat +timeZoneHrs/24.))); 
				Date date = new Date (DateTimeUtils.julDat2Millis(julDat));//+0.8/86400.0)); // add  0.5s  to avoid round errors
//				Log.d(TAG, "    getTimeZone: " +sdf.format(new Date(0)) +"  " +date.getHours()+":"+date.getMinutes()+":"+date.getSeconds()); 
				return s;
			} catch (Exception e) {
//				Log.d("PlanetDroid", "zeitStrg: " +s + "   Pattern: "+ pattern);
				return DateTimeUtils.zeitStrg (julDat, timeZoneHrs); // fallback to AstroRoutine if Error
			}
		}
	}

	public static String dayOfWeek (double julDat, double timeZoneHrs, String pattern) {
		return dayOfWeek(julDat, timeZoneHrs, pattern, Locale.getDefault());
	}
	
	/** returns Short day of Week
	 *   
	 * @param julDat
	 * @param timeZoneHrs
	 * @param pattern eg. "E" = "EEE" ("Mon"...), "EEEE" ("Monday"....), will return Kyrillic Serbian!
	 * @return
	 */
	public static String dayOfWeek (double julDat, double timeZoneHrs, String pattern, Locale locale) {
		if (locale == null) locale = Locale.US;
		SimpleDateFormat sdf= new SimpleDateFormat (pattern, locale);
		sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
		//sdf.setTimeZone(getTimeZone(timeZoneHrs));

		String dow = sdf.format(new Date(
				DateTimeUtils.julDat2Millis(AstroUtils.realMod(julDat +timeZoneHrs/24., 7.)+2455950.)	
//				millisForFormatter(
//						AstroUtils.realMod(julDat, 7.)+2455950., // set JD to 2455950 to 2455956.999... to avoid negative numbers 
//						timeZoneHrs)
				));	
		return dow.substring(0, 1).toUpperCase(locale) + dow.substring(1); 
	}

	/** returns Short day of Week
	 *   
	 * @param julDat
	 * @param timeZoneHrs
	 * @return  eg. "Mon", "Tue", "Wed",    (German: "Mo.", "Di.", "Mi.") 
	 */
	public static String dayOfWeek (double julDat, double timeZoneHrs) {
		return dayOfWeek (julDat, timeZoneHrs, "E");
	}

	/** set first letter of String in Capital (Uppercase)
	 * 
	 * @param s
	 * @return
	 */
	public static String inCapital (String s) {
		if (s != null && s.length() > 0) 
			{ return s.substring (0, 1).toUpperCase() +s.substring(1);}
		else return s;
	}
	
	/** calculates julDat from formatted String s
	 * 
	 * @param s
	 * @param timeZoneHrs
	 * @param pattern
	 * @return julDat
	 */
	public static double jdCalc (String s, double timeZoneHrs, String pattern, Locale locale) {
		if (locale == null) locale = Locale.US;
		s = s.replace("\n", "").replace("\t", " ").trim();
		if (pattern.startsWith("EEE,")) {
			s = s.substring(4).trim(); // Delete DOW from String
			pattern = pattern.replace("EEE,", "").trim(); // // Delete DOW from Pattern
		}
 	   SimpleDateFormat formatter = new SimpleDateFormat(pattern.replace("H", "k"), locale); // USes default locale
 	   formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
	   Date date;
	   double julDat;
	   try {
		   date = formatter.parse(s); 
		   julDat = DateTimeUtils.millis2JulDat(date.getTime()) 
		   -(timeZoneHrs /*-DateTimeUtils.GetCurrentTimeZone()*/)/24. 
		   +DateTimeUtils.JD_ROUND_PREVENTER;
		   //Log.d(TAG, "    Parse: " +" date: " + date.getTimezoneOffset() +" CurrTZ: "+DateTimeUtils.GetCurrentTimeZone() +"  tz: " +timeZoneHrs);
	   } catch (ParseException e) {
//		   Log.d(TAG, "Error parsing: " +s +"   pattern: " +pattern);
		   e.printStackTrace();
		   julDat = DateTimeUtils.jdCalc(s, timeZoneHrs);
	   }	 
	   //Log.d(TAG, "    returns: " +DateTimeUtils.ZeitStrg (julDat, 0) +" UTC+0");
	   return julDat; 
	}
	
	/** format value in scientifix notation, eg "12.345 e6"
	 * 
	 * @param locale
	 * @param decimals
	 * @param value
	 * @return
	 */
	public static String engFormat (Locale locale, int decimals, double value) {
		String pattern = "##0.";
		for (int i = 1; i < decimals; i++) pattern += "#";
		pattern += "E0";
		String comma = String.format(locale,  "%1.1f", 1.1).replace("1", ""); // get decimal comma character
		NumberFormat formatter = new DecimalFormat(pattern); // will generate decimal point
//		Log.d(TAG, pattern);
		return formatter.format(value).replace(".", comma).replace("E", " e"); 
	}

	/** call FormatterUtils.getBuildDate (getPackageManager().getApplicationInfo(getPackageName(), 0));
	 * 
	 * @param ai
	 * @return
	 */
	public static String getBuildDate (ApplicationInfo ai) {
		String s =  ""; //BuildConfig.COMPILE_DATE;// CompileDat.COMPILE_DAT;
		s = dateStrg(SystemUtils.getBuildMillis (ai), YMDPattern, Locale.US);
		return s;
	}
}
