package strickling.utils;

/* Orbital Elements from the internet (AstDys, JPL): 
 * see SelectPlanetoid on  package com.planetdroid
 */


/*****************************************************************************\
 *
 *  Copyright 2019 Dr. Wolfgang Strickling
 *  published under  GNU GPL 3
 *
 *   This file is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 \*****************************************************************************/
 

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.StringTokenizer;

import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.util.Log;

public class OrbElem {
	
    public long   asteroidNumber; 			//  Asteroid number (0 if unnumbered or Comet).
    public String name; 					//  Name or preliminary designation.
    public boolean isComet;					//  for calculating correct magnitude
    public double epochOfOsculation; 		//  Epoch of osculation, JulDate (TDT). If MeanAnomaly = 0 this is Time of perihelion
    public double perihelionTime; 	    	//  time of perihelion passage. If MeanAnomaly = 0 this is Time of perihelion
    public double equinoxOfElements;  		//  Equinox,    JulDate (TDT).
    public double perihelionDistance;		//  in AU 
    public double meanAnomaly; 		 		//  M0, Mean anomaly, radians
    public double argumentOfPerihelion; 	//  Omega, Argument (longitude) of perihelion, radians
    public double longitudeOfAscendingNode; //  omega, Longitude of ascending node, radians
    public double inclination; 				//  i, Inclination, radians
    public double eccentricity; 			//  e, Eccentricity.
    public double absMag; 					//  absolute Magnit.
    public double dMag; 					//  Magnitude depend. of PhaseAngle (Asteroid) or Activity param) (Comet)
    										//  Wepner 235 gives 0.023 an. in File Mean 0.15
    public double diameter; 				//  diameter [km]
 
    public static String cometURL = 
//    	"http://www.minorplanetcenter.org/iau/Ephemerides/Comets/Soft02Cmt.txt";
	    "https://minorplanetcenter.net/iau/Ephemerides/Comets/Soft02Cmt.txt";
    public static final String cometFileName0 = "comets.dat";
    public static       String cometFileName;
    public final static String SETTING_PREFIX = "Elem";
	
	public static final double JULIAN_CENTURIES = 36525.0; // Meeus p. 126  
    
	
	// constants, may be overwritten for toString routine
    public static String cometName		 			= "Comet";
    public static String asteroidName		 		= "Asteroid";
    public static String epochOfOsculationName		= "Epoch"; 		
    public static String equinoxOfElementsName		= "Equinox";  		
    public static String perihelionTimeName			= "Perihelion pass"; 	    	
    public static String perihelionDistanceName		= "Perihelion dist";		
	public static String semimajorAxisName	 		= "Semimajor Axis"; 
    public static String eccentricityName		 	= "Eccentricity";    
    public static String meanAnomalyName			= "Mean Anomaly"; 		 		
    public static String argumentOfPerihelionName	= "Argument Perih."; 	
    public static String longitudeOfAscendingNodeName = "Lon. Asc. Node"; 
    public static String inclinationName		 	= "Inclination"; 				
	public static String orbitalPeriodName		 	= "Orbital Period"; 	
    public static String absMagName		 			= "Absolute Magn."; 					
    public static String dMagName		 			= "dMag Parameter"; 					    										
    public static String diameterName		 		= "Diameter"; 				
    public static String AUName						= "AU";		
	public static String yearsName		 			= "years"; 	
    public static String magName		 			= "mag";
    public static String kilometerName		 		= "km"; 				
    

    public OrbElem (boolean istComet) {
    	this.asteroidNumber         	= 0; 		
    	this.name                   	= ""; 	
    	this.isComet                	= istComet;    
    	this.epochOfOsculation      	= 0.0; 	
    	this.perihelionTime      		= 0.0; 	
    	this.equinoxOfElements      	= 0.0;  	
    	this.perihelionDistance     	= 0.0;	
    	this.meanAnomaly            	= 0.0; 		
    	this.argumentOfPerihelion   	= 0.0;
    	this.longitudeOfAscendingNode 	= 0.0; 
    	this.inclination            	= 0.0; 	
    	this.eccentricity           	= 0.0; 	
    	this.absMag                 	= 0.0; 			
    	this.dMag                   	= 0.0;
    	this.diameter					= 0.0;
    }
    
    /**
     * 
     * @param asteroidNumber
     * @param name
     * @param isComet
     * @param epochOfOsculation
     * @param equinoxOfElements
     * @param perihelionDistance
     * @param meanAnomaly
     * @param argumentOfPerihelion
     * @param longitudeOfAscendingNode
     * @param inclination
     * @param eccentricity
     * @param absMag
     * @param dMag
     * @param diameter
     */
    public OrbElem (long asteroidNumber, String name, boolean isComet, 
    		double epochOfOsculation, double equinoxOfElements, 
    		double perihelionDistance, 
    		double meanAnomaly, double argumentOfPerihelion, 
    		double longitudeOfAscendingNode, double inclination, 
    		double eccentricity, double absMag, double dMag, double diameter) {
    	this.asteroidNumber         = asteroidNumber;
    	this.name                   = name;
    	this.isComet                = isComet;
    	this.epochOfOsculation      = epochOfOsculation;
    	this.equinoxOfElements      = equinoxOfElements;
    	this.perihelionDistance     = perihelionDistance;
    	this.meanAnomaly            = meanAnomaly;
    	this.argumentOfPerihelion   = argumentOfPerihelion;
    	this.longitudeOfAscendingNode =longitudeOfAscendingNode;
    	this.inclination            = inclination;
    	this.eccentricity           = eccentricity;
    	this.absMag                 = absMag;
    	this.dMag                   = dMag;
    	this.diameter               = diameter;
		this.perihelionTime 	    = calcPerihelionJD ();	
    }
    
    public OrbElem (OrbElem sOrbElem) {
//    	OrbElem tOrbElem = new OrbElem (sOrbElem.isComet);
		this.asteroidNumber         = sOrbElem.asteroidNumber;
    	this.name                   = sOrbElem.name;
    	this.isComet                = sOrbElem.isComet;
    	this.epochOfOsculation      = sOrbElem.epochOfOsculation;
    	this.equinoxOfElements      = sOrbElem.equinoxOfElements;
    	this.perihelionDistance     = sOrbElem.perihelionDistance;
    	this.meanAnomaly            = sOrbElem.meanAnomaly;
    	this.argumentOfPerihelion   = sOrbElem.argumentOfPerihelion;
    	this.longitudeOfAscendingNode =longitudeOfAscendingNode;
    	this.inclination            = sOrbElem.inclination;
    	this.eccentricity           = sOrbElem.eccentricity;
    	this.absMag                 = sOrbElem.absMag;
    	this.dMag                   = sOrbElem.dMag;
    	this.diameter               = sOrbElem.diameter;
		this.perihelionTime 	    = sOrbElem.perihelionTime;	
    }

    /** Converts the angles into radians
     * 
     */
    public void toRadians (){
    	this.meanAnomaly              *= Math.PI/180.0;
    	this.argumentOfPerihelion     *= Math.PI/180.0;
    	this.longitudeOfAscendingNode *= Math.PI/180.0;
    	this.inclination              *= Math.PI/180.0;
    }
    
    /*  reduce elements from equinoxOfElements to targetEquinoxTDT
     *   Meeus p. 147 ff
     */
    public OrbElem reduceElements (double targetEquinoxTDT) {
    	OrbElem tOrbElem = new OrbElem (this);
    	tOrbElem.equinoxOfElements = targetEquinoxTDT;
    	
    	final double secrad = 4.8481368111e-6; // arcsec -> radians
    	
    	double t1   = (this.equinoxOfElements - 2451545.0)/JULIAN_CENTURIES;
    	double dt   = (targetEquinoxTDT -this.equinoxOfElements)/JULIAN_CENTURIES;
    	
    	// Meeus 20.5 p 128    from https://api.kde.org/4.x-api/kdeedu-apidocs/marble/html/astrolib_8cpp_source.html
    	double eta =  ((47.0029-(0.06603-0.000598*t1)*t1) +
    		       ((-0.03302+0.000598*t1)+0.00006*dt)*dt)*dt * secrad;
    	double ppi = Math.toRadians(174.876383889) + (((3289.4789+0.60622*t1)*t1) +
    	           ((-869.8089-0.50491*t1) + 0.03536*dt)*dt)*secrad;
    	double pa = ((5029.0966+(2.22226-0.000042*t1)*t1) +
    	          ((1.11113-0.000042*t1)-0.000006*dt)*dt)*dt *secrad;
    	
    	double psi = ppi +pa;
    	
    	double i  = eta;  // for inclination = 0;
    	double sini = Math.sin(i);
    	
    	if (this.inclination == 0.0) {
    		tOrbElem.longitudeOfAscendingNode = psi + Math.PI;
    	} else {

    		// Meeus 23.1 p. 147   expressions linke sin (i) = y; cos (i) = x; => i = Math.atan2 (y, x);
    		i   = Math.acos(Math.cos(this.inclination)*Math.cos(eta) 
    				+Math.sin(this.inclination) *Math.sin(eta) *Math.cos(this.longitudeOfAscendingNode-ppi));
    		sini = Math.sin(i);

    		// Meeus 23.2 p. 147
    		double node_psi = Math.atan2(Math.sin(this.inclination) *Math.sin(this.longitudeOfAscendingNode-ppi)/sini,
    				(-Math.sin(eta)*Math.cos(this.inclination) 
    						+Math.cos(eta) *Math.sin(this.inclination) *Math.cos(this.longitudeOfAscendingNode-ppi))/sini);
    		tOrbElem.longitudeOfAscendingNode = node_psi +psi;
    	}

    	// Meeus 23.3 p. 148
    	double domega = Math.atan2(-Math.sin(eta) *Math.sin (this.longitudeOfAscendingNode-ppi)/sini,
    			(Math.sin(this.inclination)*Math.cos(eta) -Math.cos(this.inclination) *Math.sin(eta) *Math.cos(this.longitudeOfAscendingNode-ppi))/sini); 
    	tOrbElem.argumentOfPerihelion = this.argumentOfPerihelion+domega;
		tOrbElem.inclination = i;

		Log.d("PlanetDroid", "Reduce elements: " + DateTimeUtils.YMDStrg (this.equinoxOfElements, 0) +" -> "+ DateTimeUtils.YMDStrg (tOrbElem.equinoxOfElements, 0));
		Log.d("PlanetDroid", "i:    " +String.format("%10.3f -> %10.3f",  Math.toDegrees(this.inclination), Math.toDegrees(tOrbElem.inclination)));
		Log.d("PlanetDroid", "Node: " +String.format("%10.3f -> %10.3f",  Math.toDegrees(this.longitudeOfAscendingNode), Math.toDegrees(tOrbElem.longitudeOfAscendingNode)));
		Log.d("PlanetDroid", "Peri: " +String.format("%10.3f -> %10.3f",  Math.toDegrees(this.argumentOfPerihelion), Math.toDegrees(tOrbElem.argumentOfPerihelion)));
		
    	return tOrbElem;
    }

    static public double q2a (double q, double e) { // See Montenbruck p. 59
    	return q *(1.0+e)/(1.0-e*e);
    }
    
    static public double a2q (double a, double e) {
    	return a *(1.0-e*e)/(1.0+e);
    }
    
    
    public double getSemimajorAxis () {
    	return q2a (this.perihelionDistance, this.eccentricity);
    }
    
    public double getOrbitalPeriod () {
    	if (this.eccentricity >= 1.0) {return 1/0;} // return NaN for e >= 1.0; 
    	else     	{ 
    		double a =  q2a (this.perihelionDistance, this.eccentricity);
    		return Math.sqrt (a * a * a);	// Third Kepler law
    	}
    }
    
    public double calcPerihelionJD () {
    	if (Math.abs(this.eccentricity) >= 1.0 || meanAnomaly == 0.0) {
    		return epochOfOsculation;
    	}
    	else {
    		double a = getSemimajorAxis ();
    		return epochOfOsculation -meanAnomaly *Math.sqrt (a*a*a)/AstroUtils.KGAUSS;	
    	}    		
    }
    
    public double getEccentricity () {
    	return eccentricity;
    }
    
    public double getPerihelionDistance () {
    	return perihelionDistance;
    }
    
    public double getMeanAnomaly () {
    	return meanAnomaly;
    }
    
    public double getArgumentOfPerihelion () {
    	return argumentOfPerihelion;
    }
    
    public double getInclination () {
    	return inclination;
    }
    
    public double getLongitudeOfAscendingNode () {
    	return longitudeOfAscendingNode;
    }
    
    public String getName () {
    	return name;
    }
    
    /** returns with Line Breaks
     * @return
     */
    public String toString () {
    	return this.toString("\n");
    }
    
    public static void setNames (CharSequence[] namesList){
    	for (int i=0; i < namesList.length; i++) {
    		try{
    			switch (i) {
    		    case 0: cometName		 			= (String) namesList [i]; break;
    		    case 1: asteroidName		 		= (String) namesList [i]; break;
    		    case 2: epochOfOsculationName		= (String) namesList [i]; break;
    		    case 3: equinoxOfElementsName		= (String) namesList [i]; break;
    		    case 4: perihelionTimeName			= (String) namesList [i]; break;	
    		    case 5: perihelionDistanceName		= (String) namesList [i]; break;
    			case 6: semimajorAxisName	 		= (String) namesList [i]; break;
    		    case 7: eccentricityName		 	= (String) namesList [i]; break;
    		    case 8: meanAnomalyName				= (String) namesList [i]; break;	
    		    case 9: argumentOfPerihelionName	= (String) namesList [i]; break;
    		    case 10: longitudeOfAscendingNodeName = (String) namesList [i]; break;
    		    case 11: inclinationName		 	= (String) namesList [i]; break;	
    			case 12: orbitalPeriodName		 	= (String) namesList [i]; break;
    		    case 13: absMagName		 			= (String) namesList [i]; break;			
    		    case 14: dMagName		 			= (String) namesList [i]; break;			    										
    		    case 15: diameterName		 		= (String) namesList [i]; break;
    		    case 16: AUName						= (String) namesList [i]; break;
    			case 17: yearsName		 			= (String) namesList [i]; break;
    		    case 18: magName		 			= (String) namesList [i]; break;
    		    case 19: kilometerName		 		= (String) namesList [i]; break;
    			}
    		} catch (Exception e) {e.printStackTrace();}
    	}
    }
    
    public String toString (String separator) {
    	String s = asteroidName +" ";
    	if (isComet) s = cometName +" ";
    	if (asteroidNumber != 0) {
    		s += "(" +asteroidNumber+") ";
    		}
    	s += name +separator;
    	try {
    		s += String.format ("%-17s= %s", epochOfOsculationName, DateTimeUtils.zeitStrg(epochOfOsculation, 0)).replace(" - ", "-") +separator;
    		s += String.format ("%-17s= %s", equinoxOfElementsName, DateTimeUtils.zeitStrg(equinoxOfElements, 0)).replace(" - ", "-") +separator;
    		s += String.format ("%-17s= %s", perihelionTimeName,    DateTimeUtils.zeitStrg(perihelionTime,    0)).replace(" - ", "-") +separator;
    		s += String.format ("%-17s=%11.6f %s", perihelionDistanceName, perihelionDistance, AUName ) +separator;
    		if (this.eccentricity < 1.0) {
    			s += String.format ("%-17s=%11.6f %s", semimajorAxisName, getSemimajorAxis(), AUName) +separator;

    			String decimals = "6";
    			if (getOrbitalPeriod () >= 1000) decimals = "3";
    			if (getOrbitalPeriod () >= 1000000) decimals = "0";
    			s += String.format ("%-17s=%11." +decimals +"f %s", orbitalPeriodName, getOrbitalPeriod (), yearsName) +separator;
    		}
    		s += String.format ("%-17s=%11.6f",  eccentricityName, eccentricity ) +separator;
    		s += String.format ("%-17s=%11.6f\u00B0", inclinationName, Math.toDegrees ( inclination)) +separator;
    		if (this.meanAnomaly != 0.0) {
    			s += String.format ("%-17s=%11.6f\u00B0", meanAnomalyName, Math.toDegrees ( meanAnomaly)) +separator;
    		}
    		s += String.format ("%-17s=%11.6f\u00B0", argumentOfPerihelionName, Math.toDegrees ( argumentOfPerihelion)) +separator;
    		s += String.format ("%-17s=%11.6f\u00B0", longitudeOfAscendingNodeName, Math.toDegrees ( longitudeOfAscendingNode)) +separator;
    		s += String.format ("%-17s=%8.3f %s",absMagName, absMag, magName) +separator;
    		s += String.format ("%-17s=%8.3f", 	 dMagName, dMag);    	                                                            	    	
    		if (diameter > 0.0) {s += separator +String.format ("%-17s=%6.0f %s", diameterName, diameter, kilometerName);}
    	} catch (Exception e) {e.printStackTrace();}
    	return s;
}
 
    public static String settingName (int index) {
    	return SETTING_PREFIX + index +" ";
    }


	/** method for getting double values lossless from settings
	 *
	 * @param settings
	 * @param key
	 * @param defaultValue
	 * @return
	 */
	public static double getPrefsDouble (SharedPreferences settings, String key, double defaultValue) {
		return Double.longBitsToDouble(settings.getLong (key,Double.doubleToRawLongBits(defaultValue)));
	}

	/** compatibility, reads double as well as float parameter to settings
	 *
	 * @param settings
	 * @param key    (double value in key +"Double", in "key" ist float value stored)
	 * @param defaultValue
	 * @return
	 */
	public static double getPrefsDoubleCompat (SharedPreferences settings, String key, double defaultValue) {
    	if (settings.contains(key+"Double") || ! settings.contains(key)) {
    		return Double.longBitsToDouble(settings.getLong (key +"Double",Double.doubleToRawLongBits(defaultValue)));
		}  else return (double) settings.getFloat(key, (float) defaultValue);
	}

	/** method for storing double values lossless in settings
	 *
	 * @param editor
	 * @param key
	 * @param dValue
	 */
	public static void putPrefsDouble (Editor editor, String key, double dValue) {
		editor.putLong (key, Double.doubleToRawLongBits(dValue));
	}

	/** compatibility, writes double as well as float parameter to settings
	 *
	 * @param editor
	 * @param key		Writes fload value into key, double value into key +"Double"
	 * @param dValue
	 */
	public static void putPrefsDoubleCompat (Editor editor, String key, double dValue) {
		editor.putLong (key +"Double", Double.doubleToRawLongBits(dValue));
		editor.putFloat(key, 		  (float) dValue);
	}


	public void addToEditor (Editor editor, String prefix) {
    	editor.putString (prefix +"name",           name);
    	editor.putLong   (prefix +"asteroidNumber", asteroidNumber          );                         
    	editor.putBoolean(prefix +"isComet",       isComet                 );
    	// store binary value of double as long to avoid precision loss compared to float!
		putPrefsDoubleCompat(editor, prefix +"epochOfOsculation", epochOfOsculation);
		putPrefsDoubleCompat(editor, prefix +"perihelionTime", perihelionTime);

    	editor.putFloat  (prefix +"perihelionDistance", (float)       perihelionDistance      );
    	editor.putFloat  (prefix +"meanAnomaly", (float)              meanAnomaly             );                      
    	editor.putFloat  (prefix +"argumentOfPerihelion", (float)     argumentOfPerihelion    );                               
    	editor.putFloat  (prefix +"longitudeOfAscendingNode", (float) longitudeOfAscendingNode);                                   
    	editor.putFloat  (prefix +"inclination", (float)              inclination             );                      
    	editor.putFloat  (prefix +"eccentricity", (float)             eccentricity            );                       
    	editor.putFloat  (prefix +"absMag", (float)                   absMag                  );                 
    	editor.putFloat  (prefix +"dMag", (float)                     dMag                    );               
    	editor.putFloat  (prefix +"diameter", (float)                 diameter                );                   
    	editor.putLong   (prefix +"SaveTime", System.currentTimeMillis());
//		Log.d("PlanetData", String.format("OrbElem.addToEditor epochOfOsculation = %15.6f, PerL = %8.5f rad = %10.6f°",
//				epochOfOsculation, argumentOfPerihelion, Math.toDegrees(argumentOfPerihelion)));
    }

    public void addToPreferences (SharedPreferences settings, int index) {
    	SharedPreferences.Editor editor = settings.edit();
    	addToEditor (editor, index);
    	editor.commit();
    }

    public void addToEditor (Editor editor, int index) {
    	addToEditor (editor, settingName (index));
    }
    
    // usage: updateSaveTime (getSharedPreferences(PREFS_NAME, 0), elemIndex);
    public static void updateSaveTime (SharedPreferences settings, int index){
    	SharedPreferences.Editor editor = settings.edit();
    	editor.putLong  (settingName (index) +"SaveTime", System.currentTimeMillis());
    	editor.commit();
    }

    public void readFromSettings (SharedPreferences settings, String prefix) {
    	asteroidNumber             = settings.getLong (prefix +"asteroidNumber", 0);
    	name                       = settings.getString (prefix +"name", "");
    	isComet                    = settings.getBoolean (prefix +"isComet", false);

    	// Try to read the double stored values, fallback to ensure down compatibility and to avoid crash
		epochOfOsculation		   = getPrefsDoubleCompat(settings, prefix +"epochOfOsculation", 0.0);
		perihelionTime			   = getPrefsDoubleCompat(settings, prefix +"perihelionTime", 0.0);

		equinoxOfElements		   = (double) settings.getFloat (prefix +"equinoxOfElements", (float) 0.0);
		perihelionDistance         = (double) settings.getFloat (prefix +"perihelionDistance", (float) 0.0);
    	meanAnomaly                = (double) settings.getFloat (prefix +"meanAnomaly", (float) 0.0);
    	argumentOfPerihelion       = (double) settings.getFloat (prefix +"argumentOfPerihelion", (float) 0.0);
    	longitudeOfAscendingNode   = (double) settings.getFloat (prefix +"longitudeOfAscendingNode", (float) 0.0);
    	inclination                = (double) settings.getFloat (prefix +"inclination", (float) 0.0);
    	eccentricity               = (double) settings.getFloat (prefix +"eccentricity", (float) 0.0);
    	absMag                     = (double) settings.getFloat (prefix +"absMag", (float) 0.0);
    	dMag                       = (double) settings.getFloat (prefix +"dMag", (float) 0.0);
    	diameter                   = (double) settings.getFloat (prefix +"diameter", (float) 0.0);

//		Log.d("PlanetDroid", String.format("OrbElem.readFromSettings epochOfOsculation = %15.6f, PerL = %8.5f rad = %10.6f°",
//				epochOfOsculation, argumentOfPerihelion, Math.toDegrees(argumentOfPerihelion)));
    }
    
    public void readFromSettings (SharedPreferences settings, int index) {
    	readFromSettings (settings, settingName (index));	
    }
    
    //* reads CometDat from Internet
    public static boolean readCometDat () {
    	boolean result = false;
    	try {
    	    // Create a URL for the desired page
    	    URL url = new URL(cometURL);

 			File dir = new File (SystemUtils.stripFileDir(cometFileName));
			if (!dir.exists()) dir.mkdirs(); // Create Directory if necessary

	   	    // Read all the text returned by the server
     	    BufferedReader in = new BufferedReader(new InputStreamReader(url.openStream()));
            OutputStream   os = new FileOutputStream(cometFileName);  // add sdcard path
            OutputStreamWriter out = new OutputStreamWriter(os, "UTF-8");
    	    String str;
    	    while ((str = in.readLine()) != null) {
    	    	out.write(str+"\n");
    	    }
    	    out.close();
    	    in.close();
    	    result = true;
    	} catch (MalformedURLException e) {
    	} catch (IOException e) {
    	}
    	return result;    	
    }
    
    public static String getCometNameFromLine (String line){
    	return line.substring(0, 43).trim();
    }
    
    public static double getCometEpoch (String line){
    	line = line.substring(43);
    	double result = 0;
    	try {
            StringTokenizer tok 		= new StringTokenizer(line, " ");
            double day         			= Double.parseDouble(tok.nextToken());
            int month          			= Integer.parseInt(tok.nextToken());
            int year           			= Integer.parseInt(tok.nextToken());
            result = DateTimeUtils.jdCalc(year, month, day, 0.0);
    	} finally {return result;}
   }
    
    // parsing a line in Comets.dat
    public boolean parseCometLine (String line) {
    	this.isComet = false;
    	this.asteroidNumber         = 0;
    	this.name                   = getCometNameFromLine (line);
    	try {line = line.substring(43);} catch (Exception e) {}
        try {
            StringTokenizer tok 		= new StringTokenizer(line, " ");
            double day         			= Double.parseDouble(tok.nextToken());
            int month          			= Integer.parseInt(tok.nextToken());
            int year           			= Integer.parseInt(tok.nextToken());
            this.epochOfOsculation     	= DateTimeUtils.jdCalc(year, month, day, 0.0);
            this.meanAnomaly            = Math.toRadians (Double.parseDouble(tok.nextToken()));
            this.perihelionDistance     = Double.parseDouble(tok.nextToken());
            this.eccentricity           = Double.parseDouble(tok.nextToken());
            this.inclination            = Math.toRadians (Double.parseDouble(tok.nextToken()));
            this.argumentOfPerihelion   = Math.toRadians (Double.parseDouble(tok.nextToken()));
            this.longitudeOfAscendingNode = Math.toRadians (Double.parseDouble(tok.nextToken()));
            this.equinoxOfElements      = DateTimeUtils.epoch2JulDat(Double.parseDouble(tok.nextToken()));
            this.absMag                 = Double.parseDouble(tok.nextToken());
            this.dMag                   = Double.parseDouble(tok.nextToken());
            this.diameter				= 0.0;
            this.isComet = true;
    		this.perihelionTime 	    = calcPerihelionJD ();

//			Log.d("PlanetDroid", String.format("OrbElem.parseCometLine epochOfOsculation = %15.6f, PerL = %8.5f rad = %10.6f°",
//					epochOfOsculation, argumentOfPerihelion, Math.toDegrees(argumentOfPerihelion)));
    	} finally {}
    	return this.isComet;
    }
    
    
    /** Searches CometDat for desired comet. For special search, set foundPos > 1
     * 
     * @param searchStrg
     * @param foundPos   (e.g. 3 for third matching record)
     * @return
     */
    public boolean parseCometDat (String searchStrg, int foundPos) {
    	//boolean result = false;
    	int found  = 0;
    	if (foundPos <= 0) {foundPos = 1;}
    	boolean success = false;
    	searchStrg = searchStrg.toUpperCase();
    	File file = new File (cometFileName);
    	try {
    	    BufferedReader br = new BufferedReader(new FileReader(file));
    	    String line;

    	    while ((line = br.readLine()) != null && found < foundPos) {
    	    	if (line.toUpperCase().contains(searchStrg)) {
    	    		found ++;
    	    		success = parseCometLine (line);
    	    	}
    	    }
    	   br.close();     	    
    	}
    	catch (IOException e) {
    	    //You'll need to add proper error handling here
    	}
    	return found >= foundPos && success;
    }
    
    
}
