package strickling.utils;

/*****************************************************************************\
 *
 *  Copyright 2019 Dr. Wolfgang Strickling
 *  published under  GNU GPL 3
 *
 *   This file is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 \*****************************************************************************/
 

import android.util.Log;
import com.mhuss.AstroLib.MathOps;
import com.mhuss.AstroLib.NoInitException;
import com.mhuss.AstroLib.PlanetData;
import com.mhuss.AstroLib.Planets;

public class PhysEph {
	
	public static int JUPITER_SYS_II = Planets.LUNA +1;
	
	
	public static double PhaseAngle (double entf, double rPlan, double rSun, double elong) {
		double phasW = Math.acos((entf*entf +rPlan*rPlan -rSun*rSun)
                / (2 * entf * rPlan) );
	    if (elong < 0)phasW = -phasW;
	    return phasW;
	}
	
	/** return phase angle in radians
	 * 
	 * @param  pD PlanetData instance
	 * @return phase angle in radians
	 */
	public static double PhaseAngle (PlanetData pD) throws NoInitException {
//		Log.d (TAG, "ecl "+pD.getEclipticRadius());
//		Log.d (TAG, "pol "+pD.getPolarRadius());
//		Log.d (TAG, "sol "+pD.getSolarRadius());
//		Log.d (TAG, "cos " +(Sqr(pD.getEclipticRadius()) +Sqr(pD.getPolarRadius()) -Sqr (pD.getSolarRadius()))
//        / (2 * pD.getEclipticRadius () * pD.getPolarRadius ())); 
		double phasW = Math.acos((AstroUtils.sqr(pD.getEclipticRadius()) 
								 +AstroUtils.sqr(pD.getPolarRadius()) 
								 -AstroUtils.sqr (pD.getSolarRadius())) 
                / (2 * pD.getEclipticRadius () * pD.getPolarRadius ()) );
		if (Math.sin(pD.getEclipticLon () -pD.getSolarLon()) < 0) phasW = -phasW;
		return phasW;
	}
	
	/** return percent illumination of phase angle (not percent area)
	 * 
	 * @param  phaseAngle in radians
	 */
	public static double PercentIllum (Double phaseAngle) {
		return 100 -Math.abs (phaseAngle *100.0/Math.PI);
	}
	
	/** Computes angular Diamater in ArcSeconds
	 * 
	 * @param distance in AU
	 * @return diameter in ArcSeconds
	 */
	public static double DiameterSec (int planet, double distance) {
		final double [] RAD_KM =
		     {   AstroUtils.R_SUN,  // Sun
		        2439.0,  // Mercury
		        6052.0,
		        AstroUtils.R_EARTH, // Earth
		        3397.2,
		       71398.0,
		       60000.0,
		       25400.0,
		       24300.0,
		        1195.0,  // Pluto
		        AstroUtils.R_MOON,  // Moon
		     			}; // Radii in km, Montenbruck S. 166
		return DiameterSec (RAD_KM [planet], distance);
	}

	public static double DiameterSec (double diameter, double distance) {
		final double RAD2SEC = 2.75759E-3; //Umrechnungsfaktor Bogenmass-Bogensekunden
//		Log.d(TAG, planet +" " +distance);
		return RAD2SEC *diameter/distance;
	}
	
	
	public static double Magnitude (int planet, double julDat, PlanetData pD) throws NoInitException {
        double rho  = pD.getEclipticRadius(); // dist. to earth
        double radp = 1; // dist. to sun	
        double mag	= 0;
        double fv	= 0;
         
		if (planet!=0 && planet!=3)
		{   //double rade = pD.getSolarRadius();
//			elong = (rade*rade + rho*rho - radp*radp) / (2*rade*rho);
//			fv     = (radp*radp + rho*rho - rade*rade) / (2*radp*rho);
//			illum = (1.0 + fv) * 50.0;
//			elong = acos_sec(elong); 
			fv = AstroUtils.RadDeg (Math.abs(PhaseAngle (pD)));
		}
		
		if (planet!=10) radp = pD.getPolarRadius();    // dist. to sun			
		else radp = rho;

		// Magnitude Merkur .. Neptun

		if (planet==1)
		{
			mag = -0.42 + 5.0*Math.log10(radp*rho) + 0.0380*fv - 0.000273*fv*fv + 0.000002*fv*fv*fv;
		//	*dia = 6.72 / rho;
		}

		if (planet==2)
		{
			mag = -4.40 + 5.0*Math.log10(radp*rho) + 0.0009*fv + 0.000239*fv*fv - 0.00000065*fv*fv*fv;
		//	*dia = 16.82 / rho;
		}

		if (planet==4)
		{
			mag = -1.52 + 5.0*Math.log10(radp*rho) + 0.016*fv;
		//	*dia = 9.36 / rho;
		}

		if (planet==5)
		{
			mag = -9.40 + 5.0*Math.log10(radp*rho) + 0.005*fv;
		//	*dia = 184.12 / rho;
		}

		if (planet==6)
		{
			double t  = (julDat-2451545.0)/36525.0;
			double ir = 28.075216 - 0.012998*t + 0.000004*t*t;
			double nr = 169.50847 + 1.394681*t + 0.000412*t*t;
			ir=Math.toRadians(ir);
			nr=Math.toRadians(nr);
			double lambda = pD.getEclipticLon();
			double beta   = pD.getEclipticLat();
			double b 	  = Math.asin(Math.sin(beta)*Math.cos(ir) - Math.cos(beta)*Math.sin(ir)*Math.sin(lambda-nr));
			double ringmagn=-2.6*Math.sin(Math.abs(b)) + 1.25*Math.sin(b)*Math.sin(b);
			mag = -8.88 + 5.0*Math.log10(radp*rho) + 0.044*fv + ringmagn;
			//*dia = 147.64 / rho;
		}

		if (planet==7)
		{
			mag = -7.19 + 5.0*Math.log10(radp*rho);
			//*dia = 70.04 / rho;
		}

		if (planet==8)
		{
			mag = -6.87 + 5.0*Math.log10(radp*rho);
			//dia = 67.0 / rho;
		}

		if (planet==9)
		{
			mag = -1.00 + 5.0*Math.log10(radp*rho);
			//*dia = 4.14 / rho;
		}

		if (planet==10)
		{
		//if (earthposrade<0.95 || earthposrade>1.05) earthposrade=1.0;
		// mag = 0.23 + 5.0*Math.log10(earthposrade*rho) + 0.026*fv + 4.0e-9*fv*fv*fv*fv;
		mag = 0.23 + 5.0*Math.log10(rho) + 0.026*fv + 4.0e-9*fv*fv*fv*fv;
		}

		if (planet==0)
		{
			mag = -26.8 + 5.0*Math.log10(rho);
		}
		return mag;
	}
	
	
	/**
	 * 
	 * @param absMag
	 * @param dMag
	 * @param rPlan
	 * @param rSun
	 * @param entf
	 * @return
	 */
	public static double AsteroidMagnitude (double absMag,  double dMag, double rPlan, double rSun, double entf) {
		double PhasW = Math.acos ((AstroUtils.sqr(entf) +AstroUtils.sqr(rPlan) -AstroUtils.sqr(rSun))
				/ (2.0 * entf * rPlan) ); //mont 111
		//if Elong < 0 then PhasW = -PhasW;
		//Wepner 235 gibt 0.023 an. in File Mittel 0.15
		return absMag + dMag * Math.abs (Math.toDegrees(PhasW)) *0.023/0.15
		+ 5.0 * Math.log10 (rPlan * entf);
	}

	/** 
	 * 
	 * @param absMag
	 * @param dMag
	 * @param rPlan
	 * @param entf
	 * @return
	 */
	public static double CometMagnitude (double absMag, double dMag, double rPlan, double entf) {
			return absMag + 5.0 * Math.log10 (entf)
			+ 2.5 * dMag * Math.log10 (rPlan);
	}


	/** physische Sonnenephemeriden
	 * 
	 * @param julDat
	 * @param args   P0, B0, ZM in radians
	 */
	public static void  SunParams (double julDat, PlanetData pd, double[] args)
	//  nach Meissen, 14.02.98, hier alles Bogenmass, AE
	{
		//  Neigung Sonnenachse zur Ekliptik
		double   Knoten  = 75.75   * AstroUtils.RADFAKT;
		double   Neigung =  7.251  * AstroUtils.RADFAKT;

		//  Konstanten fuer P0
		double   A1      =  26.534 * AstroUtils.RADFAKT;
		double   P1      = 285.298 * AstroUtils.RADFAKT;
		double   A2      =   0.388 * AstroUtils.RADFAKT;
		double   P2      = 269.06  * AstroUtils.RADFAKT;
		double   A3      =   0.01  * AstroUtils.RADFAKT;
		double   P3      = 269.06  * AstroUtils.RADFAKT;

		double aeL, d;

		//	Rechtw3D (x, y, z, 1.0, eklL -Knoten, eklB);
		//	dreh     (y, z, Neigung);
		//	polar3D  (x, y, z, r, aeL, B0);
		double xyzVec[] = new double [3];
		double lbrVec[] = new double [3];
		try {
			MathOps.polarToCartesian (xyzVec, pd.getEclipticLon() -Knoten, pd.getEclipticLat(), pd.getEclipticRadius());
		} catch (NoInitException e) {
			e.printStackTrace();
			Log.d("PhysEph.SunParams", "NoInitException!");
		}	   
		MathOps.rotateVector (xyzVec, Neigung, 1);
		MathOps.cartesianToPolar (xyzVec,lbrVec);

		aeL = lbrVec[0] + Knoten;
		args[1] = lbrVec[1];	                     // B0

		//  Fouriertransf P0
		args[0]  =  A1 * Math.sin(    (-aeL + P1))   // P0 
		+ A2 * Math.sin(3 * (-aeL + P2))
		+ A3 * Math.sin(5 * (-aeL + P3));

		//  ZM berechnen
		d  = julDat - 2451545.0;
		args[2] = (aeL* AstroUtils.DEGFAKT - (84.11 + 14.1844 * d  + 1.222 * d/36525.0) + 165.357)/360.0;

		args[2] = args[2] - Math.floor(args[2]) + 1.0;
		if (args[2] >= 1.0) {args[2] -= 1.0;}
		args[2] *= AstroUtils.TWO_PI;
	}  //  SunParams

	
	public static int GetCarringtonNr (double julDat, PlanetData pd) {
		// liefert zu julDat die Rotationsnummer. Nach Meissen
		double  r, ZM, b, Rot;
		double [] P0_B0_ZM = new double [3];


		SunParams (julDat, pd, P0_B0_ZM);
		ZM = P0_B0_ZM [2];

		r    = (julDat - 2398167.0) / 27.2753046 + 1.0;  // grober Wert
		Rot  = Math.floor(r - Math.floor(r)   - 1 + ZM / AstroUtils.TWO_PI);
		Rot  = Math.floor(r) + Rot;
		return (int) Math.round (Math.floor (Rot + (1 - ZM / AstroUtils.TWO_PI)+0.0000001));
	}

	// ********************************************************
	// physical Ephemides
	// ********************************************************
	
    /** Calculate 
     * 
     * @param planet  Planets.SUN etc.
     * @param tdt     (JulDat - DeltaT)
     * @param pd      PlanetData
     * @param results P0, B0, ZM, if Jupiter ZM II
     * @throws NoInitException 
     */
	public static void calcP0_B0_ZM (int planet, double tdt, PlanetData pd, double [] results) throws NoInitException
	  // not for Moon!
	{
		
	  double [] [] achse =	 // Montenbruck S. 113
//	       {{  RA0, ;   dR:    AeqR:    DE0:     dD:   AeqD:  0.154},  //Sonne
	       {{ 285.96,  0.000,  0.183,  63.96,  0.000,  0.154},  //Sonne
	        { 281.02, -0.033,  0.276,  61.45, -0.005,  0.107},  // Merkur 
	        { 272.78,  0.000, -0.043,  67.21,  0.000,  0.027},  // Venus
	        { 000.00,  0.000,  0.000,  00.00,  0.000,  0.000},  // Earth, dummy line
	        {317.681, -0.108,  0.786, 52.886, -0.061,  0.413},  // Mars
	        { 268.05, -0.009,  0.116,  64.49,  0.003, -0.018},  // Jupiter
	        {  40.66, -0.036,  4.731,  83.52, -0.004,  0.407},  // Saturn
	        { 257.43,  0.000,  1.429, -15.10,  0.000, -0.114},  // Uranus
	        { 295.33,  0.000,  0.849,  40.65,  0.000,  0.242},  // Neptun
	        { 311.63,  0.000,  1.250,   4.18,  0.000,  0.374},  // Pluto                      
	        { 000.00,  0.000,  0.000,  00.00,  0.000,  0.000}}; // Moon

	                                        
	  double d, T,   //Dates
	  		 x, y, z;	
	 
	  if (planet == Planets.SUN)
	    { SunParams (tdt, pd, results);}
	  
	    else { // Merkur ... Pluto

	    AnglePair coord2000, zm_B0; 
	    // reduzieren auf 2000.0,
	    coord2000 =  AstroUtils.redEqu(pd.getRightAscension(), pd.getDeclination(), tdt, AstroUtils.JD2000);
	    coord2000 = new AnglePair (pd.getRightAscension(), pd.getDeclination());
	    d = tdt -2451545.0 -0.005776*Math.abs(pd.getEclipticRadius());
	    T = d /36525;
	    ARA = Math.toRadians (achse [planet] [0] +(achse [planet] [1] +achse [planet] [2])*T);
	    ADE = Math.toRadians (achse [planet] [3] +(achse [planet] [4] +achse [planet] [5])*T);

	    x = Math.sin (ADE) * Math.cos (coord2000.lat) - Math.cos (ADE)*Math.sin(coord2000.lat)*Math.cos(ARA -coord2000.lon);
	    y = Math.cos (ADE) * Math.sin (ARA -coord2000.lon);
	    z = Math.sqrt(1 -x*x -y*y);
	    
		double [] p_b_rVec = new double[3];
		AstroUtils.cartesianToPolar3 (x, y, z, p_b_rVec);
		results [0] = p_b_rVec [0];  // P0
	    //Polar3D (x, y, z, r, PosW, PlBrErde); // nur Absolutwert!)

	    zm_B0 = CentralMer (planet, coord2000.lon, coord2000.lat, d, T); // jetzt wird PlBrErde korrigiert
	    results [2] = zm_B0.lon;  // ZM 
//	    results [1] = Math.PI/2-zm_B0.lat;  // B0
	    results [1] = zm_B0.lat;  // B0
	    if (planet == Planets.JUPITER) { 
	    	zm_B0 = CentralMer (JUPITER_SYS_II, coord2000.lon, coord2000.lat, d, T);
		    results [3] = zm_B0.lon;  // ZM II	    	
	    }

	    }//else  Merkur ... Pluto
	} // of calcP0_B0_ZM

	
	/** calculates center of lunar disc in radians.
	 *  If eklL and eklB topocentric, then tocentric view of disc!
	 *  if eklL and eklB are heliocentric, then result is 2*PI -colongitude of the Moon 
	 *  Colongitude. is 360 deg - Longitude of morning terminator
	 * @param julDat
	 * @param eklL ecliptical longitude, radians
	 * @param eklB ecliptical latitude, radians
	 * @return Libration in Longitude and latitude
	 */
	public static AnglePair MoonLibrat (double julDat, double eklL, double eklB) 
	//see Zdenjek Kopal: The Moon (Copies p. 19)
	//Sign: East  = right (Sky West) > 0
	//West = left  (Sky East)  < 0
	{
	double AchsNeig = 1.5169 * AstroUtils.RADFAKT;  //Achsneig. i gg. Ekl = 1 Grad 32 '

	double T, AKm, x, y, z, r, l, lLib, bLib;
	
		// Kasten Montenbruck S. 103 
		T   = (julDat -2415020.0 )/36525.0;
		l   = Math.toRadians( 270.434164
				+360.0 *(1336.0*T) +307.883142*T
				-0.001133 *T*T);
		AKm = Math.toRadians( 259.183275
				-360.0 * (5.0 * T)  -134.142008*T
				+0.002078 *T*T);

		x   =  Math.cos (eklL -AKm)  * Math.cos (eklB);
		y   =  Math.sin (eklL -AKm)  * Math.cos (eklB) * Math.cos (AchsNeig)
			-Math.sin (eklB) * Math.sin (AchsNeig);
		z   = -Math.sin (eklL -AKm)  * Math.cos (eklB) * Math.sin (AchsNeig)
			-Math.sin (eklB) * Math.cos (AchsNeig);

		double [] l_b_rVec = new double[3];
		AstroUtils.cartesianToPolar3 (x, y, z, l_b_rVec);

		//writeln (raddeg (lLib):10:2, raddeg (l):10:2, raddeg (akm):10:2);

		l_b_rVec [0] = l_b_rVec [0] - l + AKm +AstroUtils.RADFAKT*0.75; //Gabel-Bug-Fummelkonstante;
		l_b_rVec [0] = AstroUtils.realMod(l_b_rVec [0], AstroUtils.TWO_PI);
//		l_b_rVec [1] = Math.PI/2 -l_b_rVec [1];
		l_b_rVec [1] = l_b_rVec [1];
		while (l_b_rVec [0] >   Math.PI) 
			{l_b_rVec [0] = l_b_rVec [0] -AstroUtils.TWO_PI;}
		while (l_b_rVec [0] < - Math.PI) 
			{l_b_rVec [0] = l_b_rVec [0] +AstroUtils.TWO_PI;}
		AnglePair result = new AnglePair (l_b_rVec [0], l_b_rVec [1]);
		return result;
	} //MoonLibrat

	public static AnglePair MoonLibrat (double julDat, PlanetData pd)
	{
		try {
			return MoonLibrat (julDat, pd.getEclipticLon (), pd.getEclipticLat());
		} catch (NoInitException e) {
			Log.d("PhysEph.MoonLibrat", "NoInitException!");
			e.printStackTrace();
			return null;
		}
	}

	/** Colongitdue of the Moon
	 * 
	 * @param julDat
	 * @param pd
	 * @return Colongitud 0..2*PI, selenocentric ecliptical latitude of the sun
	 */
	public static AnglePair MoonColong (double julDat, PlanetData pd)
	// for wax. Moon 19.4.1991 22.00 MESZ is lTermin. 23.9; lCol 336.1!!
	// Waxing Moon: Colong = 270 deg -> 90 deg, vaning Moon: 90 -> 270  deg
	// Equinoxe of date
	{
		AnglePair Colong;
		try {
			Colong = MoonLibrat (julDat, pd.getSolarLon ()-Math.PI, -pd.getSolarLat());
		} catch (NoInitException e) {
			Log.d("PhysEph.MoonColong", "NoInitException!");
			Colong = null;
			e.printStackTrace();
		}
		Colong.lon = -Colong.lon + AstroUtils.HALF_PI;
	while (Colong.lon > AstroUtils.TWO_PI) Colong.lon  -= AstroUtils.TWO_PI;
	while (Colong.lon < 0                ) Colong.lon  += AstroUtils.TWO_PI;
	return Colong;
	}  //MoonColong    

	/** Calculate Longitude of Terminator on visible lunar disc
	 * 
	 * @param colongitude
	 * @return longitude of terminator
	 */
	public static double Colong2Terminator (double colongitude)
	// Morning Terminator =  360 deg - Colong. See. Ruekl 212
	// Morning Terminator =  180 deg - Colong. See. Ruekl 212
	// Waxing Moon: Colong = 270 deg -> 90 deg, vaning Moon: 90 -> 270  deg
	{
		double terminator = 2*Math.PI -colongitude;
		if (colongitude > Math.PI /2 && colongitude < Math.PI *3/2) // Vaning Moon
			{ terminator += Math.PI; }
		while (terminator >  Math.PI /2) terminator -= Math.PI;
		while (terminator < -Math.PI /2) terminator += Math.PI;
		return terminator;
	}
	
	
	private static double ARA, ADE;

	private static AnglePair CentralMer (int planet, double ra2000, double de2000, double d, double T)  //Montenbruck S. 118  2. Formel
	{
		double [] [] meridian =                    // Montenbruck S. 114
			//   AeqW muss < 0 !!! 25.2.95 /
			//   W0 empir. korr nach Carrington-Rot, SONNE 32 S 200 27.2.95
			// W0, dW, AeqW
		{{ 84.11+0.05,14.1844000, -1.222},    // Sun,
				{329.71,      6.1385025,  1.145},    // Merkur
				{159.91,     -1.4814205,  1.436},    // Venus
				{0.0,       0.0        ,  0.000},    // Earth, dummy line
				{176.655,   350.8919830,  0.620},    // Mars
				{ 67.10,    877.9000000,  1.291},    // Jupiter System I
				{227.2037,  844.3000000, -3.470},    // Saturn
				{261.62,   -554.9130000,  0.564},    // Uranus
				{107.21,    468.7500000,  0.662},    // Neptun
				{252.66,    -56.3640000,  0.413},    // Pluto                      
				{319.50,     13.1770000,  0.000},    // Moon 
				{ 43.30,    870.2700000,  1.291}};   // Jupiter System II

		double ZM, w,  x, y, z; 

		w =  Math.toRadians (AstroUtils.realMod (meridian [planet] [0]  +meridian [planet] [1] *d +meridian [planet] [2]*T, 360));
		x =  Math.sin(de2000);
		y =  Math.cos(de2000) *Math.cos(ARA-ra2000);
		z = -Math.sin(ADE) *x -Math.cos(ADE) *y;
		y = -Math.cos(ADE) *x +Math.sin(ADE) *y;
		x =  Math.cos(de2000) * Math.sin(ARA -ra2000);

		double [] l_b_rVec = new double[3];
		AstroUtils.cartesianToPolar3 (x, y, z, l_b_rVec);
		// polar3D (x, y, z, r, ZM, PlBrErde);
		ZM = w -l_b_rVec[0];
		if (meridian [planet] [1] < 0 || planet == Planets.SUN)
		{ ZM = -ZM;}
        while (ZM < 0) {ZM = ZM +2*Math.PI;}
		while (ZM > 2*Math.PI) {ZM = ZM -2*Math.PI;}
		return new AnglePair(ZM, l_b_rVec[1]);
	} // of centralmer


	
	
}
