package strickling.utils;


/*****************************************************************************\
 *
 *  Copyright 2019 Dr. Wolfgang Strickling
 *  published under  GNU GPL 3
 *
 *   This file is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 \*****************************************************************************/
 

import com.mhuss.AstroLib.Astro;
import com.mhuss.AstroLib.AstroOps;
import com.mhuss.AstroLib.Latitude;
import com.mhuss.AstroLib.Longitude;
import com.mhuss.AstroLib.NoInitException;
import com.mhuss.AstroLib.ObsInfo;
import com.mhuss.AstroLib.Planets;
import com.mhuss.AstroLib.TimePair;

public class RiseSet extends com.mhuss.AstroLib.RiseSet {
	
	
	public static final double SUN_RISE_ALT = -0.833 * Math.PI/180.0; 
	public static final double INVALID_ANGLE = -999.9; 

	/**
	  * blue hour limit altitude (when sun is 6 degrees above horizon).
	  */
	public static final double GOLDEN_HOUR_ALT = Math.toRadians(6.);
	public static final int    GOLDEN_HOUR = 5;

	/*------------------------------------------------------------------*/
	/* Funktion MINI_MOON  from Montenbruck/Pfleger                     */
	/* Mondkoordinaten geringer Genauigkeit ( ca.5'/1' )                */
	/* RA : Rektaszension ( radians )                                   */
	/* DEC: Deklination ( radians )                                     */
	/* ( Aequinoktium des Datums )                                      */
	/*------------------------------------------------------------------*/
	void mini_moon ( double julDat, double [] ra_de )
	{
		final double M_PI2=Math.PI *2.0;
		double t = (julDat-2452545.0)/36525.0;
		double arc = 206264.8062, coseps = 0.91748; 
		double sineps = 0.39778; /* cos/sin ( Ekliptikschiefe ) */
		double l0, l, ls, f, d, h, s, n, dl, cb;
		double l_moon, b_moon, v, w, x, y, z, rho;

		/* mittlere Elemente der Mondbahn */
		/* mittl. Laenge des Mondes ( in r ) */  
		l0 = AstroUtils.frac ( 0.606433 + 1336.855225 * t ); 
		/* mittl. Anomalie des Mondes */
		l  = M_PI2 * AstroUtils.frac ( 0.374897 + 1325.552410 * t ); 
		/* mittl. Anomalie Sonne */
		ls = M_PI2 * AstroUtils.frac ( 0.993133 +  99.997361 * t ); 
		/* Diff. Laenge Mond-Sonne */
		d  = M_PI2 * AstroUtils.frac ( 0.827361 + 1236.853086 * t ); 
		/* Knotenabstand */
		f  = M_PI2 * AstroUtils.frac ( 0.259086 + 1342.227825 * t ); 
		dl = + 22640.0 * Math.sin ( l ) 
		-  4586.0 * Math.sin ( l - d - d ) 
		+  2370.0 * Math.sin ( d + d ) 
		+   769.0 * Math.sin ( l + l ) 
		-   668.0 * Math.sin ( ls ) 
		-   412.0 * Math.sin ( f + f ) 
		-   212.0 * Math.sin ( l + l - d - d ) 
		-   206.0 * Math.sin ( l + ls - d - d )
		+   192.0 * Math.sin ( l + d + d ) 
		-   165.0 * Math.sin ( ls - d - d ) 
		-   125.0 * Math.sin ( d ) 
		-   110.0 * Math.sin ( l + ls )
		+   148.0 * Math.sin ( l - ls ) 
		-    55.0 * Math.sin ( f + f - d - d );

		s = f + ( dl + 412.0 * Math.sin ( f + f ) + 541.0 * Math.sin ( ls ) ) / arc; 
		h = f - d - d;
		n = -526.0 * Math.sin ( h ) 
		+ 44.0 * Math.sin ( l + h ) 
		- 31.0 * Math.sin ( -l + h ) 
		- 23.0 * Math.sin ( ls + h ) 
		+ 11.0 * Math.sin ( -ls + h ) 
		- 25.0 * Math.sin ( -l - l + f ) 
		+ 21.0 * Math.sin ( -l + f );

		l_moon = M_PI2 * AstroUtils.frac ( l0 + dl / 1296e3 ); /* in rad */
		b_moon = ( 18520.0 * Math.sin ( s ) + n ) / arc; /* in rad */
		/* aequatoriale Koordinaten */
		cb = Math.cos ( b_moon );
		x = cb * Math.cos ( l_moon ); 
		v = cb * Math.sin ( l_moon ); 
		w = Math.sin ( b_moon );
		y = coseps * v - sineps * w; 
		z = sineps * v + coseps * w; 
		rho = Math.sqrt ( 1.0 - z * z );
		ra_de [1] = Math.atan ( z / rho ); 

		ra_de [0] = Math.atan ( y / ( x + rho ) ); 
		if ( ra_de [0] < 0 ) ra_de [0] += M_PI2;
	}

    /*-------------------------------------------------------------------------*/
    /* Funktion MINI_SUN     from Montenbruck/Pfleger                          */
    /*                                                                         */
    /* Sonnenkoordinaten geringer Genauigkeit ( ca.1' )                        */
    /* RA : Rektaszension ( radians )                                          */
    /* DEC: Deklination ( radians )                                            */
    /* ( Aequinoktium des Datums )                                    15.05.88 */
    /*-------------------------------------------------------------------------*/

    void mini_sun ( double julDat, double [] ra_de)
    {
    	final double M_PI2=Math.PI *2.0;
    	double t = (julDat-2452545.0)/36525.0;
    	final double coseps = 0.91748; 
    	final double sineps = 0.39778;
    	double l, m, dl, sl, x, y, z, rho;

    	m  = M_PI2 * AstroUtils.frac ( 0.993133 + 99.997361 * t ); 
    	dl = 6893.0 * Math.sin ( m ) + 72.0 * Math.sin ( m + m );
    	l  = M_PI2 * AstroUtils.frac ( 0.7859453 + m / M_PI2 + ( 6191.2 * t + dl ) / 1296e3 );
    	sl = Math.sin ( l );
    	x = Math.cos ( l ); 
    	y = coseps * sl; 
    	z = sineps * sl; 
    	rho = Math.sqrt ( 1.0 - z * z );
    	ra_de [1] = Math.atan ( z / rho ); 
    	ra_de [0] = Math.atan ( y / ( x + rho ) ); 
    	if ( ra_de [0] < 0 ) ra_de [0] += M_PI2;
    }

    /**
     * Calculate the rise/set (or start/end) and Culmination times for topocentric coordinates.
     *
     * @param rsType Rise/Set type (SUN, MOON, CIVIL_TWI, NAUTICAL_TWI, or ASTRONOMICAL_TWI).
     *        for other planets set <B>rsType</B> = planet +1;
     * @param jd The Julian day at 0h zone time
     * @param oi The observer's location
     * @param pd A PlanetData instance to use for our calculations.
     * @return the TimePair result. Local time, not TDT!
     */
     public static TimePair getTimesExtended( int rsType, double jd,
                                  ObsInfo oi, PlanetoidData pd) {
    	 double azimut [] = new double [3];
    	 return getTimesExtended (rsType, jd, oi, pd, azimut);
     }
     
     
     /**
      * Calculate the rise/set (or start/end) and Culmination times for topocentric coordinates.
      *
      * @param rsType Rise/Set type (SUN, MOON, CIVIL_TWI, NAUTICAL_TWI, or ASTRONOMICAL_TWI).
      *        for other planets set <B>rsType</B> = planet +1;
      * @param jd The Julian day at 0h zone time
      * @param oi The observer's location
      * @param pd A PlanetData instance to use for our calculations.
      * @param azimut  (double [3]) = azRise, asSet, azCulm
      * @return the TimePair result. Local time, not TDT!
      */
      public static TimePair getTimesExtended( int rsType, double jd,
                                   ObsInfo oi, PlanetoidData pd, double [] azimut)
     {
       //default to SUN
       double risesetAlt = SUN_ALT;  // r/s altitude
       int planet = Planets.EARTH;
       if (rsType > 10) planet = rsType -10; // if rsType <= 10 rsType is Sun, Moon, Twilights etc., if > 10 it is a planet 

       switch ( rsType ) {
       case SUN:
       case Planets.SUN+10:   
         risesetAlt = SUN_ALT;    // for topocentric Coordinate take SUN_TOPO_ALT
         break;
       case MOON:
       case Planets.LUNA+10:   
         risesetAlt = MOON_ALT;    // for topocentric Coordinate take SUN_TOPO_ALT
         planet = Planets.LUNA;    // moon
         break;
       case CIVIL_TWI:
         risesetAlt = C_TWI_ALT;
         break;
       case NAUTICAL_TWI:
         risesetAlt = N_TWI_ALT;
         break;
       case ASTRONOMICAL_TWI:
           risesetAlt = A_TWI_ALT;
           break;
       case GOLDEN_HOUR:
           risesetAlt = GOLDEN_HOUR_ALT;
           break;
       default:
     	risesetAlt = PLANET_ALT;
         break;
       }

         /*
      * Mark both the rise and set times as INVALID, to indicate that
      * they've not been found.  Note that it may turn out that one or
      * both do not occur during the given 24 hours.
      */
       TimePair riseSet = new TimePair( Astro.INVALID, Astro.INVALID, Astro.INVALID );
       azimut [0] = INVALID_ANGLE; azimut [1] = INVALID_ANGLE; azimut [2] = INVALID_ANGLE;	// mark as invalid

       double altitude[]  = new double[Astro.IHOURS_PER_DAY+1];    // 24 hrs + 1
       double hourAngle[] = new double[Astro.IHOURS_PER_DAY+1];    // 24 hrs + 1

       // Compute the altitude for each hour:
       //
       for( int i=0; i<=Astro.IHOURS_PER_DAY; i++ ) {
         pd.calc( planet, jd + AstroOps.toDays(i), 0.0, oi, false, false, false);
         try {
//           altitude[i] = Math.asin( pd.getAltAzRadius() ) - risesetAlt;
       	  // Strickling correction after correction of PlanetData return methods
       	  altitude[i]  = pd.getAltAzLat() - risesetAlt;
       	  hourAngle[i] = pd.hourAngle ();  // Culmination extension
//       	  Log.d ("RISESET", "    Hour Angle [" +i+"]  " +pd.hourAngle ());
         } catch ( NoInitException ni ) {}
       }

       // Scan the hours, looking for rise/sets:
       //
       for( int i=0; i<Astro.IHOURS_PER_DAY; i++ ) {
         int rs = UNKNOWN;
         if( altitude[i] <= 0. && altitude[i+1] > 0.) {
           // object is rising
           rs = RISING;
         }
         else if( altitude[i] > 0. && altitude[i+1] <= 0. ) {
           // object is setting
           rs= SETTING;
         }
         if( hourAngle[i] <= 0. && hourAngle[i+1] > 0.) {
             // object is culminating
             rs = CULMINATING;
           }

         if ( UNKNOWN != rs ) {
           // we found a rise or set to refine

           double fraction = AstroOps.toDays(i);
           double alt0 = altitude[i];
           double altDiff = altitude[i+1] - alt0;
           double hAng0 = hourAngle[i];
           double hAngDiff = hourAngle[i+1] - hAng0;
           double delta = 1.;
           double azi0  = INVALID_ANGLE;
           int iterations = 10;

           while( delta > .0001 && iterations > 0 ) {
             iterations--;
             if ( CULMINATING == rs && riseSet.c == Astro.INVALID) 
               {delta = ( -hAng0 / hAngDiff )/ Astro.HOURS_PER_DAY;}
             else delta = ( -alt0 / altDiff ) / Astro.HOURS_PER_DAY;
             fraction += delta;
             pd.calc( planet, jd + fraction, 0.0, oi, false, false, false); //Parameters: topo: False, Light-Time: false
             try {
//               alt0 = Math.asin( pd.getAltAzRadius() ) - risesetAlt;
              	  // Strickling correction after correction of PlanetData return methods
           	  alt0  = pd.getAltAzLat() - risesetAlt; 
           	  hAng0 = pd.hourAngle();
           	  azi0 = Math.PI -pd.getAltAzLon();
             } catch ( NoInitException ni ) {}
           }
           if      ( RISING == rs && riseSet.a == Astro.INVALID) {
        	   riseSet.a = fraction;
        	   azimut [0] = azi0;
           }
           else if ( SETTING == rs && riseSet.b == Astro.INVALID) {
        	   riseSet.b = fraction;
        	   azimut [1] = azi0;
           }
           else if ( CULMINATING == rs && riseSet.c == Astro.INVALID) {
        	   riseSet.c = fraction;
        	   azimut [2] = azi0;
           }
         }
       }
       return riseSet;
     }

	
      /**  Calculates sun rise, set an culmination times, result in JulDate, UTC, NOT TDT!
       * 
       * @param julDat in UTC= TDT -deltaT
       * @param timeZone
       * @param location
       * @param rsc Rise, Set Culmination in UTC, NOT TDT
       */
      public static void sunRiseSetCulm (double julDat, double timeZone, GeoCoordinate location, double [] rsc) {
    	  double azimut [] = new double [3]; 
    	  sunRiseSetCulm (julDat, timeZone, location, rsc, azimut);   
      }
      
      /**  Calculates sun rise, set an culmination times, result in JulDate, UTC, NOT TDT!
       * 
       * @param julDat in UTC= TDT -deltaT
       * @param timeZone
       * @param location
       * @param rsc Rise, Set Culmination in UTC, NOT TDT
       * @param azimut   Azimut (in rad) for Rise, Set, Culm
       */
      public static void sunRiseSetCulm (double julDat, double timeZone, GeoCoordinate location, double [] rsc, double [] azimut) {
    	  ObsInfo obsInfo = new ObsInfo (
				  new Latitude(location.getLat()),
				  new Longitude(location.getLon()),
				  (int) Math.round(timeZone));
		  PlanetoidData pd = new PlanetoidData(Planets.SUN, julDat, obsInfo);
		  double jdOffset = Math.floor(julDat - 0.5 + timeZone / 24.0) + 0.5 - timeZone / 24.0;

		  TimePair tRCS = getTimesExtended(com.mhuss.AstroLib.RiseSet.SUN, jdOffset, obsInfo, pd, azimut);
		  rsc[0] = tRCS.a + jdOffset;
		  rsc[1] = tRCS.b + jdOffset;
		  rsc[2] = tRCS.c + jdOffset;
    }

	/**  Calculates sun rise, set an culmination times, result in JulDate, UTC, NOT TDT!
	 *
	 * @param julDat in UTC= TDT -deltaT
	 * @param timeZone
	 * @param location
	 * @param rsc Rise, Set Culmination in UTC, NOT TDT add deltaT for obtaining TDT
	 */
	public static void moonRiseSetCulm (double julDat, double timeZone, GeoCoordinate location, double [] rsc) {
		double azimut [] = new double [3];
		moonRiseSetCulm (julDat, timeZone, location, rsc, azimut);
	}



	/**  Calculates moon rise, set an culmination times, result in JulDate, UTC, NOT TDT!
	 *
	 * @param julDat in UTC= TDT -deltaT
	 * @param timeZone
	 * @param location
	 * @param rsc Rise, Set Culmination in UTC, NOT TDT
	 * @param azimut   Azimut (in rad) for Rise, Set, Culm
	 */
	public static void moonRiseSetCulm (double julDat, double timeZone, GeoCoordinate location, double [] rsc, double [] azimut) {
		ObsInfo obsInfo = new ObsInfo (
				new Latitude(location.getLat()),
				new Longitude(location.getLon()),
				(int) Math.round(timeZone));
		PlanetoidData pd = new PlanetoidData(Planets.LUNA, julDat, obsInfo);
		double jdOffset = Math.floor(julDat - 0.5 + timeZone / 24.0) + 0.5 - timeZone / 24.0;

		TimePair tRCS = getTimesExtended(com.mhuss.AstroLib.RiseSet.MOON, jdOffset, obsInfo, pd, azimut);
		rsc[0] = tRCS.a + jdOffset;
		rsc[1] = tRCS.b + jdOffset;
		rsc[2] = tRCS.c + jdOffset;
	}

}
